#%%
import numpy as np
import matplotlib.pyplot as plt
import gpflow
from gpflow.utilities import print_summary, set_trainable

def f(x):
    return np.sin(3*x) + x

xtrain = np.linspace(0, 3, 50).reshape([-1, 1])
ytrain = f(xtrain) # + 0.5*(np.random.randn(len(xtrain)).reshape([-1, 1]) - 0.5)

ymean = np.mean(ytrain)
shift = gpflow.kernels.Constant()
shift.variance.assign(ymean**2)

rbf = gpflow.kernels.SquaredExponential()
rbf.lengthscales.assign(3.0*(np.max(xtrain) - np.min(xtrain))/len(xtrain))  # Guess not too broad, not too fine
rbf.variance.assign(3.0*np.var(ytrain))                                     # Guess from existing data range

# Alternative: if you want "real prior" data range
#rbf.variance.assign(10.0*((ytrain.max() - ytrain.min()))**2)  # Set broad or manually
#set_trainable(rbf.variance, False)  # 

k = rbf + shift
m = gpflow.models.GPR(data=(xtrain, ytrain), kernel=k)


opt = gpflow.optimizers.Scipy()

def objective_closure():
    return -m.log_marginal_likelihood()

opt_logs = opt.minimize(objective_closure,
                        m.trainable_variables,
                        options=dict(maxiter=100))

print_summary(m)


xpl = np.linspace(-5, 10, 100).reshape(100, 1)
mean, var = m.predict_y(xpl)

plt.figure(figsize=(12, 6))
plt.plot(xtrain, ytrain, 'x')
plt.plot(xpl, mean, 'C0', lw=2)
plt.fill_between(xpl[:, 0],
                 mean[:, 0] - 1.96 * np.sqrt(var[:,0]),
                 mean[:, 0] + 1.96 * np.sqrt(var[:,0]),
                 color='C0', alpha=0.2)


#%% The same with proFit
import profit

f_fit = profit.fit(xtrain, ytrain)
print_summary(f_fit.m)

mean, var = f_fit.predict(xpl)

plt.figure(figsize=(12, 6))
plt.plot(xtrain, ytrain, 'x')
plt.plot(xpl, mean, 'C0', lw=2)
plt.fill_between(xpl[:, 0],
                 mean[:, 0] - 1.96 * np.sqrt(var[:,0]),
                 mean[:, 0] + 1.96 * np.sqrt(var[:,0]),
                 color='C0', alpha=0.2)




class Postprocessor:
    def __init__(self, interface):
        self.interface = interface
        self.eval_points = read_input()

    def read(self):
        nrun = self.eval_points.shape[1]
        cwd = os.getcwd()

        self.data = np.empty(np.append(self.interface.shape(), nrun))

        # TODO move this to UQ module
#        distribution = J(*uq.params.values())
#        nodes, weights = generate_quadrature(uq.backend.order + 1, distribution, rule='G')
#        expansion = orth_ttr(uq.backend.order, distribution)

        for krun in range(nrun):
            fulldir = path.join(config.run_dir, str(krun).zfill(3)) #.zfill(3) is an option that forces krun to have 3 digits$
            try:
                os.chdir(fulldir)
                self.data[:,krun] = self.interface.get_output()
            finally:
                os.chdir(cwd)

def evaluate_postprocessing(distribution,data,expansion):
    import matplotlib.pyplot as plt
    from profit import read_input
    from chaospy import generate_quadrature, orth_ttr, fit_quadrature, E, Std, descriptives

    nodes, weights = generate_quadrature(uq.backend.order+1, distribution, rule='G')
    expansion = orth_ttr(uq.backend.order, distribution)
    approx = fit_quadrature(expansion, nodes, weights, np.mean(data[:,0,:], axis=1))
    urange = list(uq.params.values())[0].range()
    vrange = list(uq.params.values())[1].range()
    u = np.linspace(urange[0], urange[1], 100)
    v = np.linspace(vrange[0], vrange[1], 100)
    U, V = np.meshgrid(u, v)
    c = approx(U,V)

    # for 3 parameters:
    #wrange = list(uq.params.values())[2].range()
    #w = np.linspace(wrange[0], wrange[1], 100)
    #W = 0.03*np.ones(U.shape)
    #c = approx(U,V,W)

    plt.figure()
    plt.contour(U, V, c, 20)
    plt.colorbar()
    plt.scatter(config.eval_points[0,:], config.eval_points[1,:], c = np.mean(data[:,0,:], axis=1))

    plt.show()

    F0 = E(approx, distribution)
    dF = Std(approx, distribution)
    sobol1 = descriptives.sensitivity.Sens_m(approx, distribution)
    sobolt = descriptives.sensitivity.Sens_t(approx, distribution)
    sobol2 = descriptives.sensitivity.Sens_m2(approx, distribution)

    print('F = {} +- {}%'.format(F0, 100*abs(dF/F0)))
    print('1st order sensitivity indices:\n {}'.format(sobol1))
    print('Total order sensitivity indices:\n {}'.format(sobolt))
    print('2nd order sensitivity indices:\n {}'.format(sobol2))



from profit.sur.sur import Surrogate
import numpy as np
from tqdm import tqdm


class ActiveLearning:
    """Class for active learning of hyperparameters.

    For computationally expensive simulations or experiments it is crucial to get the most information out of every
    training point. This is not the case in the standard procedure of randomly selecting the training points.
    In order to get the most out of the least number of training points, the next point is inferred by
    calculating the maximum of the marginal variance, as in Garnett (2014) or Osborne (2012).
    This point then contributes the most information to the underlying function.

    Attributes:
        runner (profit.run.Runner): Runner class to dynamically start runs.
        sur (profit.sur.Surrogate): Surrogate used for fitting.
        inputs (dict): Dictionary of input points from config.
        al_keys (list): Variable names which are actively learned.
        al_ranges (list): Ranges to select the active learning variables from.
        X (ndarray): Input training data.
        y (ndarray): Observed output data.
        Xpred (ndarray): Prediction points.
        nrand (int): Number of runs with random points before active learning starts.
        ntrain (int): Total number of training runs.
        optimize_every (int): Number of active learning iterations between hyperparameter optimizations.
        plot_every (bool/int): Number of active learning iterations between plotting the progress.
            If no plots should be generated, it should be False.
        plot_marginal_variance (bool): If a subplot of the marginal variance should be included in the plots.

    Parameters:
        runner (profit.run.Runner): Runner class to dynamically start runs.
        surrogate (profit.sur.Surrogate): Surrogate used for fitting.
        inputs (dict): Dictionary of input points from config.

    Default parameters:
        nrand: 3
        optimize_every: 1
        plot_every: False
        plot_marginal_variance: False
        al_range: (0, 1)
        save: False
    """

    _defaults = {'nrand': 3, 'optimize_every': 1, 'plot_every': False, 'plot_marginal_variance': False,
                 'al_range': (0, 1), 'save': False}

    def __init__(self, runner, surrogate, inputs):
        self.runner = runner
        self.sur = surrogate
        self.inputs = inputs
        self.al_keys = [key for key in self.inputs if self.inputs[key]['kind'] == 'ActiveLearning']
        self.al_ranges = [self.inputs[key]['al_range'] for key in self.al_keys]
        self.X = self.runner.flat_input_data
        self.y = None
        self.Xpred = None
        self.nrand = None
        self.ntrain = None  # TODO: leave this open and specify a convergence criterion instead
        self.optimize_every = None
        self.plot_every = False
        self.plot_marginal_variance = False

    def run_first(self, nrand=3):
        """Runs first simulations with random points as a basis for active learning.

        The points are selected from a Halton sequence.

        Parameters:
            nrand (int): Number of runs with random points before active learning starts.
        """
        from profit.util.variable_kinds import uniform, halton
        self.nrand = self.nrand or nrand
        params_array = [{} for _ in range(self.nrand)]
        halton_seq = halton(size=(self.nrand, len(self.al_keys)))
        for idx, (key, rng) in enumerate(zip(self.al_keys, self.al_ranges)):
            for n in range(self.nrand):
                params_array[n][key] = (rng[1]-rng[0]) * halton_seq[n, idx] + rng[0]  #uniform(start=rng[0], end=rng[1], size=1)
        self.runner.spawn_array(params_array, blocking=True)
        self.X = self.runner.flat_input_data
        self.y = self.runner.flat_output_data
        try:
            self.sur.train(self.X[:self.nrand], self.y[:self.nrand], return_hess_inv=True)
        except TypeError:
            # If the surrogate does not support the advanced calculation of the marginal variance.
            self.sur.train(self.X[:self.nrand], self.y[:self.nrand])

    def update_run(self, krun, al_value):
        """Updates the input file and execute a single run.

        Parameters:
            krun (int): Current training index.
            al_value (ndarray): Selected values for the next training point.
        """
        params = {}
        for pos, key in enumerate(self.inputs):
            if key in self.al_keys:
                params[key] = al_value[pos]
                self.X[krun, pos] = al_value[pos]
        # Start single run
        self.runner.spawn_run(params, wait=True)
        # Read output (all available data)
        self.y = self.runner.flat_output_data

    @classmethod
    def from_config(cls, runner, config, base_config):
        """Instantiates an ActiveLearning object from the configuration parameters.

        Parameters:
            runner (profit.run.Runner): Runner class to dynamically start runs.
            config (dict): Only the 'active_learning' part of the base_config.
            base_config (dict): The whole configuration parameters.

        Returns:
            profit.fit.ActiveLearning: Instantiated surrogate.
        """
        sur = Surrogate.from_config(base_config['fit'], base_config)
        self = cls(runner, sur, base_config['input'])
        self.nrand = config['nrand']
        self.ntrain = base_config['ntrain']
        self.optimize_every = config['optimize_every']
        self.plot_every = config['plot_every']
        self.plot_marginal_variance = config['plot_marginal_variance']
        self.Xpred = config.get('Xpred')
        return self

    @classmethod
    def handle_config(cls, config, base_config):
        """Sets default values in the configuration if the parameter is not existent.

        Parameters:
            config (dict): Only the 'active_learning' part of the base_config.
            base_config (dict): The whole configuration parameters.
        """
        if config is None:
            config = {'ActiveLearning': {}}

        for key, default in cls._defaults.items():
            if key not in config:
                config[key] = default

        for key, value in base_config['input'].items():
            if value['kind'] == 'ActiveLearning':
                if not value.get('al_range'):
                    value['al_range'] = cls._defaults['al_range']
        if config.get('save'):
            from os import path
            config['save'] = path.join(base_config['base_dir'], config['save'])

    @staticmethod
    def f(u):
        r"""Example function for debugging.

        This function is used as a reference that is plotted beside the fit and training data points.

        $$
        \begin{equation}
        f(u) = cos(10u) + u
        \end{equation}
        $$
        """
        return np.cos(10 * u) + u

    def learn(self, ntrain=None, Xpred=None, optimize_every=1, plot_every=False, plot_marginal_variance=False):
        """Main loop for active learning.

        Parameters:
            ntrain (int): Total number of training runs.
            Xpred (ndarray): Prediction points.
            optimize_every (int): Number of active learning iterations between hyperparameter optimizations.
            plot_every (int): Number of active learning iterations between plotting the progress.
            plot_marginal_variance (bool): If a subplot of the marginal variance should be included in the plots.
        """

        # Set variables either from config or from given parameter
        self.ntrain = self.ntrain or ntrain
        self.Xpred = self.Xpred or Xpred or self.sur.default_Xpred()
        self.optimize_every = self.optimize_every or optimize_every
        self.plot_every = self.plot_every or plot_every
        self.plot_marginal_variance = self.plot_marginal_variance or plot_marginal_variance
        if self.plot_every:
            from matplotlib.pyplot import subplots, show

        # Create a dense predictive array for each dimension. This becomes inefficient for high dimensions.
        if isinstance(self.Xpred, list):
            xp = [np.arange(minv, maxv, step) for minv, maxv, step in self.Xpred]
            self.Xpred = np.hstack([xi.flatten().reshape(-1, 1) for xi in np.meshgrid(*xp)])

        # Plot first runs
        if self.plot_every:
            def create_fig():
                subplot_kw = {'projection': '3d'} if self.Xpred.shape[-1] > 1 else None
                fig, ax = subplots(1 + self.plot_marginal_variance, 1,
                                   subplot_kw=subplot_kw)
                if not self.plot_marginal_variance:
                    ax = [ax]
                return ax

            ax = create_fig()
            self.sur.plot(self.Xpred, ref=self.f, axes=ax[0])

        # Main loop
        for krun in tqdm(range(self.nrand, self.ntrain)):
            try:
                marginal_variance = self.sur.get_marginal_variance(self.Xpred)
            except AttributeError:
                # Surrogate must have marginal variance calculation
                raise RuntimeError("Surrogate {} is not suited for Active Learning!".format(self.sur.__class__.__name__))

            # Find next candidate
            if np.max(marginal_variance.max(axis=0) - marginal_variance.min(axis=0)) >= 1e-5:
                # Normalized marginal variance plus a penalty for near points.
                loss = marginal_variance / marginal_variance.max() + \
                       self.additional_loss(self.X[krun - 1].reshape(1, -1))
                loss /= loss.max()

                # Plot marginal variance in a subplot
                if self.plot_every and not (krun+1) % self.plot_every and self.plot_marginal_variance:
                    if self.Xpred.shape[-1] > 1:
                        ax[1].plot_trisurf(self.Xpred[:, 0], self.Xpred[:, 1], loss.flatten())
                    else:
                        ax[1].plot(self.Xpred, loss)

                # Next candidate
                al_value = self.Xpred[np.argmax(loss)]
            else:
                # If marginal variance is not expressive, randomly select the next training point.
                al_value = self.Xpred[np.random.randint(self.Xpred.shape[0])]
            self.update_run(krun, al_value)
            self.sur.add_training_data(self.X[krun].reshape(1, -1), self.y[krun].reshape(1, -1))

            # Optimize hyperparameters
            if not (krun+1) % self.optimize_every:
                self.sur.optimize(return_hess_inv=True)

            # Plot progress
            if self.plot_every and not (krun+1) % self.plot_every:
                ax = create_fig()
                self.sur.plot(self.Xpred, ref=self.f, axes=ax[0])

        if self.plot_every:
            show()

    def additional_loss(self, last_point):
        r"""Penalty for candidates which are near to the previous point.

        $$
        \begin{equation}
        L = 1 - \exp(-\frac{1}{2} \lvert X_{pred} - X_{last} \rvert)
        \end{equation}
        $$

        Parameters:
            last_point (ndarray): Last training point.

        Returns:
            float: Penalty between 0 and 1.
        """

        return 1.0 - np.exp(-0.5 * np.linalg.norm(self.Xpred - last_point, axis=1).reshape(-1, 1))

    def save(self, path):
        """Saves the surrogate model.

        Parameters:
            path (str): Path including the file name, where the model should be saved.
        """

        self.sur.save_model(path)



import numpy as np

from profit.util import halton
from profit.sur.backend.gpfunc import *

ndim = 1
ntrain = 10

xtrain = np.asfortranarray(halton.halton(ntrain, ndim))
K = np.empty((ntrain, ntrain), order='F')

print(K.shape)
print(xtrain.shape)
print(ntrain)

gpfunc.build_k_sqexp(xtrain, xtrain, K)



"""
Testcases for some configuration features:
- .yaml and .py configuration files
- text file in run directory from template
- text and json format in template
- hdf5 input
- relative symbolic links
- default config values
"""

from profit.config import Config
from profit.util import load
from json import load as jload
from os import path, remove, chdir, getcwd
from subprocess import run
from numpy import ndarray, genfromtxt, array
from shutil import rmtree
from pytest import fixture


@fixture(autouse=True)
def chdir_pytest():
    pytest_root_dir = getcwd()
    chdir(path.dirname(path.abspath(__file__)))
    yield
    chdir(pytest_root_dir)


TIMEOUT = 30  # seconds


def clean(config):
    """Delete run directories and input/output files after the test."""

    for krun in range(config.get('ntrain')):
        single_dir = path.join(config.get('run_dir'), f'run_{krun:03d}')
        if path.exists(single_dir):
            rmtree(single_dir)
    if path.exists('./study/interface.npy'):
        remove('./study/interface.npy')
    if path.exists(config['files'].get('input')):
        remove(config['files'].get('input'))
    if path.exists(config['files'].get('output')):
        remove(config['files'].get('output'))
    if path.exists(config['run'].get('log_path')):
        rmtree(config['run'].get('log_path'))


def test_yaml_py_config():
    """Tests if .yaml and .py configuration files are equal by comparing dict keys and values."""

    yaml_file = '././study/profit.yaml'
    py_file = '././study/profit_config.py'
    config_yaml = Config.from_file(yaml_file)
    config_py = Config.from_file(py_file)

    def assert_dict(dict_items1, dict_items2):
        for (key1, value1), (key2, value2) in zip(dict_items1, dict_items2):
            assert key1 == key2
            if type(value1) is dict:
                assert_dict(value1.items(), value2.items())
            elif type(value1) is ndarray:
                assert value1.dtype == value2.dtype
                assert value1.shape == value2.shape
            elif key1 != 'config_path':
                assert value1 == value2

    assert_dict(config_yaml.items(), config_py.items())


def test_txt_input():
    """Tests if the input files in the single run directories are created from the template."""

    config_file = './study/profit.yaml'
    config = Config.from_file(config_file)
    run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
    assert path.isfile('./study/run_000/mockup.in')
    clean(config)


def test_txt_json_input():
    """Checks if the numpy arrays resulting from a text and a json input are equal."""

    config_file = './study/profit_json.yaml'
    config = Config.from_file(config_file)
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        with open(path.join(config['run_dir'], 'run_000', 'mockup_json.in')) as jf:
            json_input = jload(jf)
        json_input = array([float(val) for val in json_input.values()])
        with open(path.join(config['run_dir'], 'run_000', 'mockup.in')) as tf:
            txt_input = genfromtxt(tf)
        assert json_input.dtype == txt_input.dtype
        assert json_input.shape == txt_input.shape
    finally:
        clean(config)


def test_hdf5_input_output():
    """Checks the data inside a .hdf5 input file."""

    config_file = './study/profit_hdf5.yaml'
    config = Config.from_file(config_file)
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        data_in = load(config['files'].get('input'))
        assert data_in.shape == (2, 1)
        assert data_in.dtype.names == ('u', 'v', 'w')
    finally:
        clean(config)


def test_symlinks():
    """Checks if relative symbolic links are handled correctly."""

    config_file = './study/profit_symlink.yaml'
    config = Config.from_file(config_file)
    base_file = './study/run_000/mockup.in'
    link_file = './study/run_000/some_subdir/symlink_link.txt'
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        with open(link_file, 'r') as link:
            with open(base_file, 'r') as base:
                link_data = link.read()
                base_data = base.read()
                assert link_data == base_data and not link_data.startswith('{')
    finally:
        clean(config)


def test_default_values():
    """ Tests the default values of the configuration file. First with a simple configuration
    and then with some parameters customized, to check if missing dict entries are set correctly. """

    # First with simple configuration
    config_file = './study/profit_default.yaml'
    config = Config.from_file(config_file)
    assert config.get('base_dir') == path.abspath('./study')
    assert config.get('run_dir') == config.get('base_dir')
    assert config['files'].get('input') == path.join(config.get('base_dir'), 'input.txt')
    assert config['files'].get('output') == path.join(config.get('base_dir'), 'output.txt')
    assert config['fit'].get('surrogate') == 'GPy'
    assert config['fit'].get('kernel') == 'RBF'

    # Now check when dicts are only partially set
    config_file = './study/profit_default_2.yaml'
    config = Config.from_file(config_file)
    assert config['files'].get('input') == path.join(config.get('base_dir'), 'custom_input.in')
    assert config['files'].get('output') == path.join(config.get('base_dir'), 'output.txt')
    assert config['fit'].get('surrogate') == 'GPy'
    assert config['fit'].get('kernel') == 'RBF'
    assert config['fit'].get('plot') is True



# %%
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

# See https://debuggercafe.com/getting-started-with-variational-autoencoder-using-pytorch/
class LinearVAE(nn.Module):
    def __init__(self, D, d):
        super(LinearVAE, self).__init__()

        self.enc = nn.Linear(in_features=D, out_features=d*2)
        self.dec = nn.Linear(in_features=d, out_features=D)

    def reparameterize(self, mu, log_var):
        std = torch.exp(0.5*log_var)
        eps = torch.randn_like(std)
        sample = mu + (eps * std)
        return sample

    def forward(self, x):
        x = self.enc(x).view(-1, 2, d)

        mu = x[:,0,:]
        log_var = x[:,1,:]
        z = self.reparameterize(mu, log_var)

        x = self.dec(x)
        return x, mu, log_var

D = 3
d = 2

Wlift = torch.Tensor([[1.0, 0], [0, 1.0], [0.0, 0.0]])

x = Wlift @ torch.rand(d)

ae = LinearVAE(D, d)
print(list(ae.parameters()))
y = ae.forward(x)
print(x)
print(y)
# %%
ntrain = 100
input = torch.rand(ntrain, D) - 0.5
output = ae(input)
target = input
criterion = nn.MSELoss()
loss = criterion(output, target)

print(loss)
#%% Torch LBFGS
# import torch.optim as optim

# # create your optimizer
# optimizer = optim.LBFGS(ae.parameters())

# for inp in input:
#     def closure():
#         if torch.is_grad_enabled():
#             optimizer.zero_grad()
#         output = ae(inp)
#         loss = criterion(output, inp)
#         if loss.requires_grad:
#             loss.backward()
#         return loss
#     optimizer.step(closure)

# output = ae(input)
# print(criterion(output, input))

# %%

print(ae(input)-input)

# %%
from scipy.optimize import minimize

def target(w):
    with torch.no_grad():
        ae.W.copy_(torch.Tensor(w.reshape([D,D])))
    if ae.W.grad is not None:
        ae.W.grad.zero_()
    output = ae(input)
    loss = criterion(output, input)
    if loss.requires_grad:
        loss.backward()

    return loss.detach().flatten().numpy(), ae.W.grad.detach().flatten().numpy()

target(ae.W)
# %%
minimize(target, ae.W.detach().flatten(), method='Powell', jac=True)

# %%



"""proFit main script.

This script is called when running the `profit` command.
"""

from os import getcwd
import sys
from argparse import ArgumentParser, RawTextHelpFormatter
import logging

from profit.config import Config
from profit.util import safe_path_to_file, safe_str

from profit.run import Runner

yes = False  # always answer 'y'


def fill_uq(self, krun, content):
    params_fill = SafeDict()
    kp = 0
    for item in self.params:
        params_fill[item] = self.eval_points[kp, krun]
        kp = kp+1
    return content.format_map(params_fill)


def main():
    """
    Main command line interface
    sys.argv is an array whose values are the entered series of command
    (e.g.: sys.argv=['profit','run', '--active-learning', '/home/user/example'])
    """

    """ Get parameters from argv """
    parser = ArgumentParser(usage='profit <mode> (base-dir)',
                            description="Probabilistic Response Model Fitting with Interactive Tools",
                            formatter_class=RawTextHelpFormatter)
    parser.add_argument('mode',  # ToDo: subparsers?
                        metavar='mode',
                        choices=['run', 'fit', 'ui', 'clean'],
                        help='run ... start simulation runs \n'
                             'fit ... fit data with Gaussian Process \n'
                             'ui ... visualise results \n'
                             'clean ... remove run directories and input/output files')
    parser.add_argument('base_dir',
                        metavar='base-dir',
                        help='path to config file (default: current working directory)',
                        default=getcwd(), nargs='?')
    args = parser.parse_args()

    print(args)

    """ Instantiate Config class from the given file """
    config_file = safe_path_to_file(args.base_dir, default='profit.yaml')
    config = Config.from_file(config_file)

    sys.path.append(config['base_dir'])

    if args.mode == 'run':
        from tqdm import tqdm
        from profit.pre import get_eval_points, write_input
        from profit.util import save

        runner = Runner.from_config(config['run'], config)

        eval_points = get_eval_points(config)
        write_input(config['files']['input'], eval_points)

        if 'activelearning' in (safe_str(v['kind']) for v in config['input'].values()):
            from profit.fit import ActiveLearning
            from profit.sur.sur import Surrogate
            runner.fill(eval_points)
            if 'active_learning' not in config:
                config['active_learning'] = {}
            ActiveLearning.handle_config(config['active_learning'], config)
            al = ActiveLearning.from_config(runner, config['active_learning'], config)
            try:
                al.run_first()
                al.learn()
            finally:
                runner.cancel_all()
            if config['fit'].get('save'):
                al.save(config['fit']['save'])
        else:
            params_array = [row[0] for row in eval_points]
            try:
                runner.spawn_array(tqdm(params_array), blocking=True)
            finally:
                runner.cancel_all()

        if config['run']['clean']:
            runner.clean()

        if config['files']['output'].endswith('.txt'):
            data = runner.structured_output_data
            save(config['files']['output'], data.reshape(data.size, 1))
        else:
            save(config['files']['output'], runner.output_data)

    elif args.mode == 'fit':
        from numpy import arange, hstack, meshgrid
        from profit.util import load
        from profit.sur.sur import Surrogate

        sur = Surrogate.from_config(config['fit'], config)

        if not sur.trained:
            x = load(config['files']['input'])
            y = load(config['files']['output'])
            x = hstack([x[key] for key in x.dtype.names])
            y = hstack([y[key] for key in y.dtype.names])

            sur.train(x, y)

        if config['fit'].get('save'):
            sur.save_model(config['fit']['save'])
        if config['fit'].get('plot'):
            try:
                xpred = [arange(minv, maxv, step) for minv, maxv, step in config['fit']['plot'].get('Xpred')]
                xpred = hstack([xi.flatten().reshape(-1, 1) for xi in meshgrid(*xpred)])
            except AttributeError:
                xpred = None
            sur.plot(xpred, independent=config['independent'], show=True)

    elif args.mode == 'ui':
        from profit.ui import init_app
        app = init_app(config)
        app.run_server(debug=True)

    elif args.mode == 'clean':
        from shutil import rmtree
        from os import path, remove
        run_dir = config['run_dir']

        question = "Are you sure you want to remove the run directories in {} " \
                   "and input/output files? (y/N) ".format(config['run_dir'])
        if yes:
            print(question + 'y')
        else:
            answer = input(question)
            if not answer.lower().startswith('y'):
                print('exit...')
                sys.exit()

        for krun in range(config['ntrain']):
            single_run_dir = path.join(run_dir, f'run_{krun:03d}')
            if path.exists(single_run_dir):
                rmtree(single_run_dir)
        if path.exists(config['files']['input']):
            remove(config['files']['input'])
        if path.exists(config['files']['output']):
            remove(config['files']['output'])

        runner = Runner.from_config(config['run'], config)
        runner.clean()
        try:
            rmtree(config['run']['log_path'])
        except FileNotFoundError:
            pass


if __name__ == '__main__':
    main()



import numpy as np
from func import (integrate_pendulum)
from param import  Nm, N, dtsymp, qmin, qmax, pmin, pmax
from profit.util.halton import halton


# specify initial conditions (q0, p0)
method = 'all'
# samples = bluenoise(2, N, method, qmin, qmax, pmin, pmax)
samples = halton(N, 2)*np.array([qmax-qmin, pmax-pmin]) + np.array([qmin, pmin])

q = samples[:,0]
p = samples[:,1]*1.5

t0 = 0.0  # starting time
t1 = dtsymp*Nm # end time
t = np.linspace(t0,t1,Nm) # integration points in time

#integrate pendulum to provide data
ysint = integrate_pendulum(q, p, t)

#%%
P = np.empty((N))
Q = np.empty((N))
for ik in range(0, N):    
    P[ik] = ysint[ik][1,-1]
    Q[ik] = ysint[ik][0,-1]
    
zqtrain = Q - q
zptrain = p - P

xtrain = q.flatten()
ytrain = P.flatten()
ztrain1 = zptrain.flatten()
ztrain2 = zqtrain.flatten()
ztrain = np.concatenate((ztrain1, ztrain2))



import numpy as np


def f(u):
    return np.cos(10 * u) + 0.1*np.random.randn(), np.sin(10 * u) + u


params = np.loadtxt('mockup_multi_output.in')
result = f(params)
print(result)
np.savetxt('mockup.out', np.array([result]))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb  4 08:34:52 2019

@author: Christopher Albert
"""
import numpy as np
import json
from h5py import File


def n_F(E, B, T):
    kb = 8.6e-5  # in eV / K
    mu = 5  # eV
    T = (10000 - 100) * T + 100  # rescale because Halton sequence is between 0 and 1
    return 1 / (np.exp((E-mu) / (B*kb*T)) + 1)


with open('mockup_independent.json') as fin:
    params = json.load(fin)
E = np.arange(0, 10, 0.1)
result = n_F(E, **params)
print(result)
with File('mockup.hdf5', 'w') as fout:
    fout['n_f'] = result



#%%
import numpy as np
import plotly.express as px
import pandas as pd
from sklearn import model_selection
import GPy

nskip = 1000
skip = np.arange(1e6)
skip = np.delete(skip, np.arange(0, 1e6, nskip))

data = pd.read_table(
    '/mnt/c/Users/chral/Dropbox/ipp/paper_algae/mc_out.dat', sep='\s+', skiprows=skip)
names = {
    'mu_0': 'k_alg_growth_max',
    'f_si': 'frac_si_alg_1',
    'lambda_S': 'k_att_shade',
    'K_light': 'k_light_sm',
    'sigma_0': 'k_alg_loss',
    'a': 'coeff_d_loss_2'
}
# %%
Ndim = 6
indata = data[names.values()].values
indata = (indata - np.min(indata,0))/np.max(indata - np.min(indata,0),0)
px.scatter(x=indata[:,0], y=indata[:,1])
outdata = data[['cost_function']].values
outdata = (outdata - np.min(outdata,0))/np.max(outdata - np.min(outdata,0),0)

xtrain, xtest = model_selection.train_test_split(indata)
ytrain, ytest = model_selection.train_test_split(outdata)
#%%
kernel = GPy.kern.RBF(input_dim=Ndim, variance=1.0, ARD=True, lengthscale=np.ones(Ndim))
m = GPy.models.GPRegression(xtrain, ytrain, kernel, noise_var=1e-4)
m.optimize(messages=True)
m
# %%
E, std = m.predict(xtrain)
px.scatter(x=ytrain.flatten(), y=E.flatten())

# %%
E, std = m.predict(xtest)
px.scatter(x=ytest.flatten(), y=E.flatten())
# %%



# Configuration file for RedMod scripts v0.1
from redmod import uq

# Backend for uncertainty quantification
uq.backend = uq.ChaosPy(order = 3, sparse = False)

# Define parameters' uncertainties
uq.read_params('params_1998.txt')



#!/usr/bin/env python
# coding: utf-8

import numpy as np

from profit.sur.gaussian_process import GPySurrogate


def f(x):
    return np.sin(x)

nx = 100
train_every = 10
x = np.linspace(0, 5, nx).reshape([nx, 1])
y = f(x)
xtrain = x[::train_every]
ytrain = f(xtrain)
nxtrain = len(xtrain)

a = np.array([1.0, 1.0])

def test_sur():
    gps = GPySurrogate()
    gps.train(x, y)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 22 00:21:20 2020

@author: manal
"""

import numpy as np
# from GPy import *
import GPy
from GPy.kern.src.stationary import Stationary
import matplotlib.pyplot as plt

class Cosin(Stationary):
    """
    Cosine kernel:
    .. math::
       k(r) = \sigma^2 \cos ( r )

       r(x, x') = \\sqrt{ \\sum_{q=1}^Q \\frac{(x_q - x'_q)^2}{\ell_q^2} }
    """
    def __init__(self, input_dim, variance=1., lengthscale=None, ARD=False, active_dims=None, name='Cosin'):
        super(Cosin, self).__init__(input_dim, variance, lengthscale, ARD, active_dims, name)

    def K_of_r(self, r):
        return self.variance * np.cos(r)

    def dK_dr(self, r):
        return -self.variance * np.sin(r)

    def dK_dX(self, X, X2, dimX):
        r = self._scaled_dist(X, X2)
        id = np.where(r<=1e-6)
        dK_dr = self.dK_dr(r)
        dist = X[:,None,dimX]-X2[None,:,dimX]
        lengthscale2inv = (np.ones((X.shape[1]))/(self.lengthscale**2))[dimX]
        k = 2. * lengthscale2inv * dist * r**(-1) * dK_dr
        for i in range(id[0].shape[0]):
            k[i,i]= -dist[i,i]
        return k
        # return r,2. * lengthscale2inv * dist * r1 * dK_dr

    def dK_dX2(self,X,X2,dimX2):
        return -1.*self.dK_dX(X,X2, dimX2)

    def dK2_dXdX2(self, X, X2, dimX, dimX2):
        if X2 is None:
            X2 = X
        r = self._scaled_dist(X, X2)
        dist = X[:,None,:]-X2[None,:,:]
        lengthscale2inv = np.ones((X.shape[1]))/(self.lengthscale**2)
        K = self.K_of_r(r)
        dK_dr = self.dK_dr(r)
        dist = X[:,None,:]-X2[None,:,:]
        lengthscale2inv = np.ones((X.shape[1]))/(self.lengthscale**2)
        l1 = lengthscale2inv[dimX]
        l2 = lengthscale2inv[dimX2]
        d1 = dist[:,:,dimX]
        d2 = dist[:,:,dimX]
        id = np.where(r<=1e-6)
        k = (dimX==dimX2)*(-2.)*l1*d1*r**(-1)*dK_dr + l1*d1*r**(-1) * (2.*l1*d1*r**(-1)*dK_dr + 4*l2*d2*K)
        for i in range(id[0].shape[0]):
            k[i,i]= -dist[i,i,dimX2]*dist[i,i,dimX]
        return k

#%%
k0 = Cosin(1, active_dims=0)
k01 = Cosin(1)
k1 = Cosin(1, active_dims=1)
k2 = Cosin(3)

x0train = np.linspace(-5,5,100).reshape(-1,1)
x1train = np.linspace(-5,5,100).reshape(-1,1)
xtrain = np.hstack((x0train, x1train))

# print('Training points:')
# print(xtrain)
# print()



k0_K = k0.K(x0train,x0train)
dk0 = k0.dK_dX(x0train, x0train, 0)
dk0_2 = k0.dK2_dXdX2(x0train, x0train, 0, 0)

# print('K0 = ')
# print(k0_K,'\n')
# print('dK0.dX: \n', dk0,'\n')
# print('dK2.dXdX2: \n',dk0_2)
# # print()
# # k0.dK_dX(x0train, x0train, 0).plot()
# plt.figure()
# plt.plot(k0_K[0,:])
# plt.show
# plt.figure()
# plt.plot(dk0[0,:])
# plt.show
# plt.figure()
# plt.plot(dk0_2[0,:])
# plt.show
# # plt.plot(r,k0.dK_dX(x0train,x0train,0))
# plt.figure()
# plt.imshow(k)
# plt.figure()
# plt.imshow(k[1:5,1:5])
# plt.figure()
# plt.imshow(r)

# k01_K = k01.K(x0train,x0train)
# dk01 = k01.dK_dX(x0train, x0train, 0)
# dk01_2 = k01.dK2_dXdX2(x0train, x0train, 0, 0)
# print('k: \n', k01_K-k0_K)
# print('dk: \n', dk01-dk0)
# print('dk2: \n', dk01_2-dk0_2)

# k1_K = k1.K(xtrain,xtrain)
# dk1 = k1.dK_dX(xtrain, xtrain, 1)
# dk1_2 = k1.dK2_dXdX2(xtrain, xtrain, 1, 1)
# print('K1 = ')
# print(k1_K,'\n')  # Need 2D vectors, here, as active_dims=1
# print('dk1.dX: \n',dk1,'\n')
# print('dk12.dXdX2: \n',dk1_2)
# plt.figure()
# plt.plot(k1_K[0,:])
# plt.show
# plt.figure()
# plt.plot(dk1[0,:])
# plt.show
# plt.figure()
# plt.plot(dk1_2[0,:])
# plt.show
# print()
# # plt.plot(k1.dK_dX)
# print('k0_der: \n', k0_der.K(xtrain,xtrain),'\n')

# xtrain2 = np.hstack((x0train, x1train, x1train))
# k2_K = k2.K(xtrain2,xtrain2)
# dk2 = k2.dK_dX(xtrain2, xtrain2, 2)
# dk2_2 = k2.dK2_dXdX2(xtrain2, xtrain2, 2, 2)
# print('K2 = ')
# print(k2_K,'\n')  # Need 2D vectors, here, as active_dims=1
# print('dk2.dX: \n',dk2,'\n')
# print('dk22.dXdX2: \n',dk2_2)
# plt.figure()
# plt.plot(k2_K[0,:])
# plt.show
# plt.figure()
# plt.plot(dk2[0,:])
# plt.show
# plt.figure()
# plt.plot(dk2_2[0,:])
# plt.show


#%%

import pytest
# from katharina import kernels
import func

def test_kern_Cosine():
    x0train = np.linspace(-5,5,100).reshape(-1,1)
    # x1train = np.linspace(-5,5,100).reshape(-1,1)
    # xtrain = np.hstack((x0train, x1train))

    k0 = Cosin(1, active_dims=0)
    k0_K = k0.K(x0train,x0train)
    dk0 = k0.dK_dX(x0train, x0train, 0)
    dk0_2 = k0.dK2_dXdX2(x0train, x0train, 0, 0)

    l = np.array([1, 1])

    K = np.zeros((len(x0train),len(x0train)))
    for i in range(K.shape[0]):
        for j in range(K.shape[1]):
            K[i,j] = func.f_kern(x0train[i], x0train[i], x0train[j], x0train[j], l)

    dK = np.zeros((len(x0train),len(x0train)))
    for i in range(dK.shape[0]):
        for j in range(dK.shape[1]):
            dK[i,j] = func.dkdx(x0train[i], x0train[i], x0train[j], x0train[j], l)

    dK2 = np.zeros((len(x0train),len(x0train)))
    for i in range(dK2.shape[0]):
        for j in range(dK2.shape[1]):
            dK2[i,j] = func.d2kdxdx0(x0train[i], x0train[i], x0train[j], x0train[j], l)

    # print(k)
    # print(",,,,")
    # print(k1a)

    plt.figure()
    plt.imshow(K)
    plt.figure()
    plt.imshow(k0_K)
    plt.figure()
    plt.plot(K[0,:])

    plt.figure()
    plt.imshow(dK)
    plt.figure()
    plt.imshow(dk0)
    plt.figure()
    plt.plot(dK[0,:])

    plt.figure()
    plt.imshow(dK2)
    plt.figure()
    plt.imshow(dk0_2)
    plt.figure()
    plt.plot(dK2[0,:])

    print(k0_K)
    print(K)
    print()
    print(dk0)
    print(dK)
    print()
    print(dk0_2)
    print(dK2)
    print()
    # print(np.isclose(k0_K, K, rtol=1e-6))
    # assert np.isclose(dk0, dK, rtol=1e-6)
    # assert np.isclose(dk0_2, dK2, rtol=1e-6)

# %%
test_kern_Cosine()



import plotly.graph_objects as go
import numpy as np


def dens_hist(a, edges, outside):
    count, edges = np.histogram(a, edges)
    if outside[0]:
        count = np.append(len(a[a < edges[0]]), count)
        edges = np.append(edges[0] - (edges[1]-edges[0]), edges)
    if outside[1]:
        count = np.append(count, len(a[a > edges[-1]]))
    density = count*100/len(a)
    return density, edges

def format_hist(fig, bins, edges, density, title, outside, colors):
    fig.update_yaxes(autorange='reversed')
    fig.update_layout(
        title=dict(text='<b>{}</b>'.format(title),
                   x=0.5, y=0.97, xanchor='center', yanchor='top'),
        titlefont=dict(size=30),
        autosize=False,
        width=380,
        height=400,
        xaxis=dict(
            linecolor='black',
            linewidth=2,
            mirror=True,
            showgrid=False,
            showticklabels=False,
            zeroline=False,
            domain=[0.15, 1]
        ),
        yaxis=dict(
            linecolor='black',
            linewidth=2,
            mirror=True,
            showgrid=False,
            showticklabels=False,
            zeroline=False,
        ),
        margin=dict(l=210, r=1, t=50, b=2),
        plot_bgcolor=colors[0],
        paper_bgcolor=colors[1],
        barmode='overlay',
        showlegend=False
    )
    annotations = []

    scale = np.max(np.abs(edges))

    if scale < 1e-2:
        fmt = '{:.2e}'
    elif scale < 1e-1:
        fmt = '{:.3f}'
    elif scale < 1e0:
        fmt = '{:.2f}'
    elif scale < 1e1:
        fmt = '{:.1f}'
    elif scale < 1e3:
        fmt = '{:.0f}'
    else:
        fmt = '{:.2e}'

    for k, xd in enumerate(density):
        yd = edges[k]
        if outside[0] and k == 0:
            label = '< {}'.format(fmt).format(bins[0])
        elif outside[1] and k == len(density)-1:
            label = '> {}'.format(fmt).format(bins[-1])
        else:
            label = '{0} - {0}'.format(fmt).format(edges[k], edges[k+1])
        annotations.append(dict(xref='paper', yref='y',
                                x=-1.2, y=yd,
                                xanchor='left',
                                text='{}'.format(label),
                                font=dict(size=24),
                                showarrow=False, align='left'))
        # labeling the bar net worth
        annotations.append(dict(xref='paper', yref='y1',
                                y=yd, x=0.14,
                                xanchor='right',
                                text='{:.1f}'.format(xd) + '%',
                                font=dict(size=24),
                                showarrow=False, align='right'))
    fig.update_layout(annotations=annotations)

def draw_hist(edges, density):
    fig = go.Figure(data=go.Bar(
            y=edges, x=density, orientation='h',
            marker=dict(color='black', line=dict(color='white', width=1))))

    return fig


def fig_hist(da, bins, title, condi=None, outside=(False, True),
             colors=('#8ecbad', '#b3ffd9')):
    density, edges = dens_hist(da, bins, outside)

    if condi is None:
        fig = draw_hist(edges, density)
    else:
        density2, edges2 = dens_hist(da[condi], bins, outside)

        bar1 = go.Bar(
            y=edges2, x=density2, orientation='h',
            marker=dict(color='black', line=dict(color='black'))
        )
        bar2 = go.Bar(
            y=edges, x=density, orientation='h',
            marker=dict(
                color='rgba(0,0,0,0)',
                line=dict(color='white', width=1)
            )
        )
        fig = go.Figure(data=[bar1, bar2])

        density = density2
        edges = edges2
    
    format_hist(fig, bins, edges, density, title, outside, colors)

    return fig


def add_border(fig):
    fig.add_shape(
        type='rect',
        xref='paper',
        yref='paper',
        x0=-0.325,
        y0=-0.125,
        x1=1.023,
        y1=1.16,
        line=dict(
            color='black',
            width=2,
        ),
    )



"""proFit worker class & components

:author: Robert Babin
:date: Mar 2021
"""

import os
import shutil
import logging
from abc import ABC, abstractmethod  # Abstract Base Class
import time
import subprocess
from typing import Mapping, MutableMapping
from numpy import zeros, void
from warnings import warn


def checkenv(name):  # ToDo: move to utils? Modify logger name
    if os.getenv(name) is None:
        logging.getLogger(__name__).error(f'{name} is not set')
        # exit(1)
        raise RuntimeError
    return os.getenv(name)


# === Interface === #


class Interface(ABC):
    interfaces = {}  # ToDo: rename to registry?
    # ToDo: inherit from a Subworker class? __init__, registry, register is basically the same for all 3

    def __init__(self, config, run_id: int, *, logger_parent: logging.Logger = None):
        self.config = config  # only the run/interface config dict (after processing defaults)
        self.logger = logging.getLogger('Interface')
        if logger_parent is not None:
            self.logger.parent = logger_parent
        self.run_id = run_id

        if 'time' not in self.__dir__():
            self.time: int = 0
        if 'input' not in self.__dir__():
            self.input: void = zeros(1, dtype=[])[0]
        if 'output' not in self.__dir__():
            self.output: void = zeros(1, dtype=[])[0]

    @abstractmethod
    def done(self):
        pass

    @classmethod
    def register(cls, label):
        def decorator(interface):
            if label in cls.interfaces:
                warn(f'registering duplicate label {label} for Interface')
            cls.interfaces[label] = interface
            return interface
        return decorator

    def __class_getitem__(cls, item):
        return cls.interfaces[item]


# === Pre === #


class Preprocessor(ABC):
    preprocessors = {}

    def __init__(self, config, *, logger_parent: logging.Logger = None):
        self.config = config  # ~base_config['run']['post']
        self.logger = logging.getLogger('Preprocessor')
        if logger_parent is not None:
            self.logger.parent = logger_parent

    @abstractmethod
    def pre(self, data: Mapping, run_dir: str):
        if os.path.exists(run_dir):
            shutil.rmtree(run_dir)
        os.mkdir(run_dir)
        os.chdir(run_dir)

    def post(self, run_dir: str, clean=True):
        os.chdir('..')
        if clean:
            shutil.rmtree(run_dir)

    def __call__(self, data: Mapping, run_dir: str):
        return self.pre(data, run_dir)

    @classmethod
    def handle_config(cls, config, base_config):
        pass

    @classmethod
    def register(cls, label):
        def decorator(pre):
            if label in cls.preprocessors:
                warn(f'registering duplicate label {label} for Preprocessor')
            cls.preprocessors[label] = pre
            return pre
        return decorator

    def __class_getitem__(cls, item):
        return cls.preprocessors[item]

    @classmethod
    def wrap(cls, label):
        def decorator(func):
            @cls.register(label)
            class WrappedPreprocessor(cls):
                def pre(self, data, run_dir):
                    func(data)
        return decorator


# === Post === #


class Postprocessor(ABC):
    postprocessors = {}

    def __init__(self, config, *, logger_parent: logging.Logger = None):
        self.config = config  # ~base_config['run']['post']
        self.logger = logging.getLogger('Postprocessor')
        if logger_parent is not None:
            self.logger.parent = logger_parent

    @abstractmethod
    def post(self, data: MutableMapping):
        pass

    def __call__(self, data: MutableMapping):
        return self.post(data)

    @classmethod
    def handle_config(cls, config, base_config):
        pass

    @classmethod
    def register(cls, label):
        def decorator(post):
            if label in cls.postprocessors:
                warn(f'registering duplicate label {label} for Postprocessor')
            cls.postprocessors[label] = post
            return post
        return decorator

    def __class_getitem__(cls, item):
        return cls.postprocessors[item]

    @classmethod
    def wrap(cls, label):
        def decorator(func):
            @cls.register(label)
            class WrappedPostprocessor(cls):
                def post(self, data):
                    func(data)
        return decorator


# === Worker === #


class Worker:
    _registry = {}

    def __init__(self, config: Mapping, interface_class, pre_class, post_class, run_id: int):
        self.logger = logging.getLogger('Worker')
        self.logger.setLevel(logging.DEBUG)
        try:
            os.mkdir(config['log_path'])
        except FileExistsError:
            pass
        log_handler = logging.FileHandler(os.path.join(config['log_path'], f'run_{run_id:03d}.log'), mode='w')
        log_formatter = logging.Formatter('{asctime} {levelname:8s} {name}: {message}', style='{')
        log_handler.setFormatter(log_formatter)
        self.logger.addHandler(log_handler)

        self.config: Mapping = config  # ~ base_config['run']
        self.run_id: int = run_id
        self.run_dir: str = f'run_{self.run_id:03d}'

        self.pre: Preprocessor = pre_class(config['pre'], logger_parent=self.logger)
        self.post: Postprocessor = post_class(config['post'], logger_parent=self.logger)
        self.interface: Interface = interface_class(config['interface'], run_id, logger_parent=self.logger)

    @classmethod
    def from_config(cls, config, run_id):
        interface = Interface[config['interface']['class']]
        pre = Preprocessor[config['pre']['class']]
        post = Postprocessor[config['post']['class']]
        return cls[config['worker']](config, interface, pre, post, run_id)

    @classmethod
    def from_env(cls, label='run'):
        from profit.config import Config
        base_config = Config.from_file(checkenv('PROFIT_CONFIG_PATH'))
        config = base_config[label]
        if config['custom']:
            cls.handle_config(config, base_config)
        run_id = int(checkenv('PROFIT_RUN_ID')) + int(os.environ.get('PROFIT_ARRAY_ID', 0))
        return cls.from_config(config, run_id)

    @classmethod
    def handle_config(cls, config, base_config):
        """ handle the config dict, fill in defaults & delegate to the children

        :param config: dict read from config file, ~base_config['run'] usually

        ToDo: check for correct types & valid paths
        ToDo: give warnings
        """
        defaults = {'pre': 'template', 'post': 'json', 'command': './simulation', 'stdout': 'stdout', 'stderr': None,
                    'clean': True, 'time': True, 'log_path': 'log'}
        for key, default in defaults.items():
            if key not in config:
                config[key] = default

        if not os.path.isabs(config['log_path']):
            config['log_path'] = os.path.abspath(os.path.join(base_config['base_dir'], config['log_path']))
        if not isinstance(config['pre'], MutableMapping):
            config['pre'] = {'class': config['pre']}
        Preprocessor[config['pre']['class']].handle_config(config['pre'], base_config)
        if not isinstance(config['post'], MutableMapping):
            config['post'] = {'class': config['post']}
        Postprocessor[config['post']['class']].handle_config(config['post'], base_config)

    @classmethod
    def register(cls, label):
        def decorator(worker):
            if label in cls._registry:
                warn(f'registering duplicate label {label} for Worker')
            cls._registry[label] = worker
            return worker
        return decorator

    def __class_getitem__(cls, item):
        if item is None:
            return cls
        return cls._registry[item]

    @classmethod
    def wrap(cls, label, inputs, outputs):
        def decorator(func):
            @cls.register(label)
            class WrappedWorker(cls):
                def main(self):
                    values = func(*[self.interface.input[key] for key in inputs])
                    if isinstance(outputs, str):
                        self.interface.output[outputs] = values
                    else:
                        for value, key in zip(values, outputs):
                            self.interface.output[key] = value
                    self.interface.done()
        return decorator

    def run(self):
        kwargs = {}
        if self.config['stdout'] is not None:
            kwargs['stdout'] = open(self.config['stdout'], 'w')
        if self.config['stderr'] is not None:
            kwargs['stderr'] = open(self.config['stderr'], 'w')
        self.logger.debug(f"run `{self.config['command']}` {kwargs if kwargs != {} else ''}")
        cp = subprocess.run(self.config['command'], shell=True, text=True, **kwargs)
        self.logger.debug(f"run returned {cp.returncode}")

    def main(self):
        self.pre(self.interface.input, self.run_dir)

        timestamp = time.time()
        self.run()
        if self.config['time']:
            self.interface.time = int(time.time() - timestamp)

        self.post(self.interface.output)
        self.interface.done()
        self.pre.post(self.run_dir, self.config['clean'])


# === Entry Point === #


def main():
    """
    entry point to run a worker

    configuration with environment variables and the main config file
    intended to be called by the Runner class
    """
    worker = Worker.from_env()
    worker.main()



"""
Testcases for mockup simulations:
- all input variables are Halton sequences to ensure reproducibility
- 1D function
    - .hdf5 input/output
    - multi output
- 2D function (Rosenbrock)
    - .txt input/output
- 2D function (Fermi)
    - one Halton (Temperature) and one independent (Energy) variable (simulation returns a vector)
    - .json template
    - .hdf5 input/output
    - .hdf5 output of single runs with custom interface
"""

from profit.config import Config
from profit.sur import Surrogate
from profit.util import load
from os import path, remove, chdir, getcwd
from subprocess import run
from numpy import array, allclose
from shutil import rmtree
from pytest import fixture


@fixture(autouse=True)
def chdir_pytest():
    pytest_root_dir = getcwd()
    chdir(path.dirname(path.abspath(__file__)))
    yield
    chdir(pytest_root_dir)


# Allow a large range of parameters, just ensure that it is approximately at the same scale
NLL_ATOL = 1e3
PARAM_RTOL = 2
TIMEOUT = 30  # seconds


def clean(config):
    """Delete run directories and input/outpt files after the test."""

    for krun in range(config.get('ntrain')):
        single_dir = path.join(config.get('run_dir'), f'run_{krun:03d}')
        if path.exists(single_dir):
            rmtree(single_dir)
    if path.exists(config['run']['interface'].get('path')):
        remove(config['run']['interface'].get('path'))
    if path.exists(config['files'].get('input')):
        remove(config['files'].get('input'))
    if path.exists(config['files'].get('output')):
        remove(config['files'].get('output'))
    if path.exists(config['run'].get('log_path')):
        rmtree(config['run'].get('log_path'))


def test_1D():
    """Test a simple function f(u) = cos(10*u) + u."""

    config_file = 'study_1D/profit_1D.yaml'
    config = Config.from_file(config_file)
    model_file = config['fit'].get('save')
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        run(f"profit fit {config_file}", shell=True, timeout=TIMEOUT)
        sur = Surrogate.load_model(model_file)
        assert sur.get_label() == 'GPy'
        assert sur.trained
        assert sur.model.kern.name == 'rbf'
        assert allclose(sur.model.likelihood.variance[0], 4.809421284738159e-11, atol=NLL_ATOL)
        assert allclose(sur.model.kern.variance[0], 1.6945780226638725, rtol=PARAM_RTOL)
        assert allclose(sur.model.kern.lengthscale, 0.22392982500520792, rtol=PARAM_RTOL)
    finally:
        clean(config)
        if path.exists(model_file):
            remove(model_file)


def multi_test_1d(study, config_file, output_file):
    """ test 1D with different config files """
    config_file = path.join(study, config_file)
    output_file = path.join(study, output_file)
    config = Config.from_file(config_file)
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        output = load(output_file)
        assert output.shape == (7, 1)
        assert all(output['f'] - array([0.7836, -0.5511, 1.0966, 0.4403, 1.6244, -0.4455, 0.0941]).reshape((7, 1))
                   < 1e-4)
    finally:
        clean(config)


def test_custom_post1():
    """ test 1D with custom postprocessor """
    multi_test_1d('./study_custom_post1', 'profit_custom_post1.yaml', 'output_custom_post1.hdf5')


def test_custom_post2():
    """ test 1D with custom postprocessor (wrap syntax) """
    multi_test_1d('./study_custom_post2', 'profit_custom_post2.yaml', 'output_custom_post2.hdf5')


def test_custom_worker1():
    """ test 1D with custom worker (custom entry point) """
    multi_test_1d('./study_custom_worker1', 'profit_custom_worker1.yaml', 'output_custom_worker1.hdf5')


def test_custom_worker2():
    """ test 1D with custom worker (integrated, custom main) """
    multi_test_1d('./study_custom_worker2', 'profit_custom_worker2.yaml', 'output_custom_worker2.hdf5')


def test_custom_worker3():
    """ test 1D with custom worker (integrated, custom run) """
    multi_test_1d('./study_custom_worker3', 'profit_custom_worker3.yaml', 'output_custom_worker3.hdf5')


def test_custom_worker4():
    """ test 1D with custom worker (integrated, custom run) """
    multi_test_1d('./study_custom_worker4', 'profit_custom_worker4.yaml', 'output_custom_worker4.hdf5')


def test_multi_output():
    """Test a 1D function with two outputs."""
    config_file = 'study_multi_output/profit_multi_output.yaml'
    config = Config.from_file(config_file)
    model_file = config['fit'].get('save')
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        run(f"profit fit {config_file}", shell=True, timeout=TIMEOUT)
        sur = Surrogate.load_model(model_file)
        assert sur.get_label() == 'GPy'
        assert sur.trained
        assert sur.model.kern.name == 'ICM'
        assert allclose(sur.model.likelihood.likelihoods_list[0].variance[0], 0.00032075301845035454, atol=NLL_ATOL)
        assert allclose(sur.model.likelihood.likelihoods_list[0].variance[0], 3.773865299540149e-09, atol=NLL_ATOL)
        assert allclose(sur.model.kern.rbf.variance[0], 0.52218353, rtol=PARAM_RTOL)
        assert allclose(sur.model.kern.rbf.lengthscale, 0.20184872, rtol=PARAM_RTOL)
    finally:
        clean(config)
        from os.path import splitext
        # .hdf5 is not yet supported for multi output model, so it is saved as .pkl instead.
        model_file = splitext(model_file)[0] + '.pkl'
        if path.exists(model_file):
            remove(model_file)


def test_2D():
    """Test a Rosenbrock 2D function with two random inputs."""

    config_file = 'study_2D/profit_2D.yaml'
    config = Config.from_file(config_file)
    model_file = config['fit'].get('save')
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        run(f"profit fit {config_file}", shell=True, timeout=TIMEOUT)
        sur = Surrogate.load_model(model_file)
        assert sur.get_label() == 'GPy'
        assert sur.trained
        assert sur.model.kern.name == 'rbf'
        assert sur.model.kern.input_dim == 2
        assert allclose(sur.model.likelihood.variance[0], 2.657441549034709e-08, atol=NLL_ATOL)
        assert allclose(sur.model.kern.variance[0], 270.2197671669302, rtol=PARAM_RTOL)
        assert allclose(sur.model.kern.lengthscale[0], 1.079943283873971, rtol=PARAM_RTOL)
    finally:
        clean(config)
        if path.exists(model_file):
            remove(model_file)


def test_2D_independent():
    """Test a Fermi function which returns a vector over energy and is sampled over different temperatures."""

    config_file = 'study_independent/profit_independent.yaml'
    config = Config.from_file(config_file)
    model_file = config['fit'].get('save')
    try:
        run(f"profit run {config_file}", shell=True, timeout=TIMEOUT)
        run(f"profit fit {config_file}", shell=True, timeout=TIMEOUT)
        sur = Surrogate.load_model(model_file)
        assert sur.get_label() == 'GPy'
        assert sur.trained
        assert sur.model.kern.name == 'rbf'
        assert sur.model.kern.input_dim == 1
        assert allclose(sur.model.likelihood.variance[0], 2.8769632382230903e-05, atol=NLL_ATOL)
        assert allclose(sur.model.kern.variance[0], 0.4382486018781694, rtol=PARAM_RTOL)
        assert allclose(sur.model.kern.lengthscale[0], 0.24077767526116695, rtol=PARAM_RTOL)
    finally:
        clean(config)
        if path.exists(model_file):
            remove(model_file)



"""Script to initialize kernels via SymPy.

Running this script generates Fortran code for kernels
that is written to `kernels_base.f90`. This code has to
be compiled via `make` subsequently.
"""
#%%
from sympy import symbols, sqrt, exp, diff
from sympy.utilities.codegen import codegen
import shutil

#%% # Kernel functions (kernel, 1st and second derivatives)
r, ra, rb = symbols(r'r r_a r_b', real=True)
l = symbols(r'l', positive=True)

kern = {}
kern['sqexp'] = exp(-r**2/2)
kern['matern32'] = (1 + sqrt(3)*r)*exp(-sqrt(3)*r)
kern['matern52'] = (1 + sqrt(5)*r + 5*r/3)*exp(-sqrt(5)*r)
kern['wend4'] = (1 - r)**4*(1 + 4*r)

dkern = {}
for k, v in kern.items():
    dkern[k] = diff(v, r).simplify()

d2kern = {}
for k, v in dkern.items():
    d2kern[k] = diff(v, r).simplify()

funlist = []
for k in kern:
    funlist = funlist + [
        ('kern_{}'.format(k), kern[k]),
        ('dkern_{}'.format(k), dkern[k]),
        ('d2kern_{}'.format(k), d2kern[k])
    ]

[(name, code), (h_name, header)] = codegen(
    funlist, 'F95', 'kernels_base', header=False, empty=False)

with open(name, 'w') as f:
    f.write(code)



"""Tests for Fortran UQ code from Roland Preuss
"""
import os
import pytest
import numpy as np
from time import time
from profit import uq


@pytest.fixture(scope='module')
def uqp():
    from fffi import FortranModule

    cwd = os.path.dirname(__file__)
    os.chdir(cwd)

    fort_mod = FortranModule('uqp', 'mod_unqu')

    fort_mod.fdef("""
      integer :: np, nall, npar, nt, iflag_run, iflag_mod, iflag_pol

      subroutine allocate_params end
      subroutine init_uq end

      subroutine set_legendre_borders(kpar, lower, upper)
        integer, intent(in) :: kpar
        double precision, intent(in) :: lower, upper
      end

      subroutine set_hermite_mean_std(kpar, mean0, std)
        integer, intent(in) :: kpar
        double precision, intent(in) :: mean0, std
      end

      subroutine pre_uq(axi)
        double precision, intent(inout) :: axi(:,:)
      end

      subroutine run_uq end
      """)

    fort_mod.compile(verbose=1)
    fort_mod.load()
    return fort_mod


@pytest.mark.skip(reason="Fortran code for UQ currently unavailable")
def test_hermite_points(uqp):
    """
    Test Hermite quadrature points in 2D
    """

    print('=== Hermite 2D quadrature points ===')

    mean1 = 5.0
    std1 = 0.1
    mean2 = 2.0
    std2 = 0.2

    t = time()  # UQP timing
    uqp.npar = 2
    uqp.allocate_params()
    uqp.np = 3
    uqp.nt = 1
    uqp.iflag_run = 1
    uqp.iflag_mod = 2
    uqp.iflag_pol = 2  # 1: Legende, 2: Hermite, 3: Laguerre
    uqp.set_hermite_mean_std(1, mean1, std1)
    uqp.set_hermite_mean_std(2, mean2, std2)
    uqp.init_uq()
    axi = np.zeros((uqp.nall, uqp.npar), order='F')
    uqp.pre_uq(axi)
    print('UQP: {:4.2f} ms'.format(1000*(time() - t)))

    t = time()  # ChaosPy timing
    uqu = uq.UQ()
    uqu.backend = uq.backend.ChaosPy(order=3, sparse=False)
    uqu.params['u'] = uqu.backend.Normal(mean1, std1)
    uqu.params['v'] = uqu.backend.Normal(mean2, std2)
    axi2 = uqu.get_eval_points().T
    print('ChaosPy: {:4.2f} ms'.format(1000*(time() - t)))

    axis = axi[np.lexsort((axi[:, 0], axi[:, 1]))]
    axi2s = axi2[np.lexsort((axi2[:, 0], axi2[:, 1]))]

    assert np.allclose(axis, axi2s)
    print('Success: UQP and ChaosPy results match')



from os import path, getcwd
from re import match, split
import yaml
from collections import OrderedDict
from typing import Mapping

from profit.run import Runner
from profit.sur import Surrogate

VALID_FORMATS = ('.yaml', '.py')

"""
yaml has to be configured to represent OrderedDict 
see https://stackoverflow.com/questions/16782112/can-pyyaml-dump-dict-items-in-non-alphabetical-order
and https://stackoverflow.com/questions/5121931/in-python-how-can-you-load-yaml-mappings-as-ordereddicts
"""


def represent_ordereddict(dumper, data):
    value = []

    for item_key, item_value in data.items():
        node_key = dumper.represent_data(item_key)
        node_value = dumper.represent_data(item_value)

        value.append((node_key, node_value))

    return yaml.nodes.MappingNode(u'tag:yaml.org,2002:map', value)


def try_parse(s):
    funcs = [int, float]
    for f in funcs:
        try:
            return f(s)
        except ValueError:
            pass
    return s


def dict_constructor(loader, node):
    return OrderedDict(loader.construct_pairs(node))


_mapping_tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG
yaml.add_representer(OrderedDict, represent_ordereddict)
yaml.add_constructor(_mapping_tag, dict_constructor)

""" now yaml is configured to handle OrderedDict input and output """


def load_config_from_py(filename):
    """ Load the configuration parameters from a python file into dict. """
    from importlib.util import spec_from_file_location, module_from_spec
    spec = spec_from_file_location('f', filename)
    f = module_from_spec(spec)
    spec.loader.exec_module(f)
    return {name: value for name, value in f.__dict__.items() if not name.startswith('_')}


class Config(OrderedDict):
    """
    Configuration class
    This class provides a dictionary with possible configuration parameters for
    simulation, fitting and uncertainty quantification.

    Possible parameters in .yaml:

    base_dir: .
    run_dir: .
    uq: # TODO: implement
    interface: ./interface.py
    files:
        input: ./input.txt
        output: ./output.txt
    ntrain: 30
    variables:
        input1:
            kind: Normal
            range: (0, 1)
            dtype: float
        ...
        independent1:
            kind: Independent
            range: (0, 10, 1)
            dtype: int
        ...
        output1:
            kind: Output
            range: independent1
            dtype: float
    run:
        cmd: python3 ../simulation.py
        ntask: 4
    fit:
        surrogate: GPy
        kernel: RBF
        sigma_n: None
        sigma_f: 1e-6
        save: ./model.hdf5
        load: ./model.hdf5
        plot: Bool
            xpred: ((0, 1, 0.01), (0, 10, 0.1))
        plot_searching_phase: Bool
    """

    def __init__(self, base_dir=getcwd(), **entries):
        super(Config, self).__init__()
        self['base_dir'] = path.abspath(base_dir)
        self['run_dir'] = self['base_dir']
        self['uq'] = {}
        self['variables'] = {}
        self['fit'] = {'surrogate': 'GPy',
                       'kernel': 'RBF'}
        self['files'] = {'input': path.join(self['base_dir'], 'input.txt'),
                         'output': path.join(self['base_dir'], 'output.txt')}

        # Not to fill directly in file
        self['independent'] = {}
        self['input'] = {}
        self['output'] = {}
        self.update(entries)

    def write_yaml(self, filename='profit.yaml'):
        """ Dump UQ configuration to a yaml file.
        The default filename is profit.yaml
        """
        dumpdict = dict(self)
        self._remove_nones(dumpdict)
        with open(filename,'w') as file:
            yaml.dump(dumpdict,file,default_flow_style=False)

    @classmethod
    def from_file(cls, filename='profit.yaml'):
        """ Load configuration from .yaml or .py file.
        The default filename is profit.yaml """
        from profit.util import variable_kinds, safe_str, get_class_methods

        self = cls(base_dir=path.split(filename)[0])

        if filename.endswith('.yaml'):
            with open(filename) as f:
                entries = yaml.safe_load(f)
        elif filename.endswith('.py'):
            entries = load_config_from_py(filename)
        else:
            raise TypeError("Not supported file extension .{} for config file.\n"
                            "Valid file formats: {}".format(filename.split('.')[-1], VALID_FORMATS))
        self.update(entries)

        if path.isabs(filename):
            self['config_path'] = filename
        else:
            self['config_path'] = path.abspath(path.join(getcwd(), filename))

        """ Variable configuration
        kind: Independent, Uniform, etc.
        range: (start, end, step=1) or {'dependent variable': (start, end, step=1)} for output
        dtype: float64
        """

        halton_dim = []
        for k, v in self['variables'].items():
            if isinstance(v, str):
                # match word(comma seperated list of int or float)
                parsed = split('[()]', v)
                kind = parsed[0]
                args = parsed[1] if len(parsed) >= 2 else ''
                entries = tuple(try_parse(a) for a in args.split(',')) if args != '' else tuple()

                if isinstance(try_parse(v), (int, float)):  # PyYaml has problems parsing scientific notation
                    kind = 'Constant'
                    entries = (try_parse(v),)

                self['variables'][k] = {'kind': kind}

                if safe_str(kind) == 'output':
                    spl = split('[()]', v)
                    if len(spl) >= 3:
                        dependent = [var.strip() for var in split(',', spl[1])]
                    else:
                        dependent = []
                    self['variables'][k]['depend'] = tuple(dependent)
                    self['variables'][k]['range'] = {k: None for k in dependent}
                else:
                    try:
                        func = getattr(variable_kinds, safe_str(kind))
                        if safe_str(kind) == 'halton':
                            halton_dim.append((k, entries))
                        elif safe_str(kind) == 'independent':
                            self['variables'][k]['range'] = func(*entries, size=self['ntrain']) if entries else None
                        elif safe_str(kind) == 'activelearning':
                            self['variables'][k]['range'] = func(size=self['ntrain'])  # to insert nan in input file
                            self['variables'][k]['al_range'] = entries  # range between active learning should happen
                        else:
                            self['variables'][k]['range'] = func(*entries, size=self['ntrain'])
                            self['variables'][k]['dtype'] = str(self['variables'][k]['range'].dtype)
                    except AttributeError:
                        raise RuntimeError(f"Variable kind {kind} not defined.\n"
                                           f"Valid Functions: {get_class_methods(variable_kinds)}")
            elif isinstance(v, (int, float)):
                self['variables'][k] = {'kind': 'constant', 'range': variable_kinds.constant(v, size=self['ntrain'])}
                self['variables'][k]['dtype'] = str(self['variables'][k]['range'].dtype)
            elif isinstance(v, Mapping) and 'kind' in v:  # assume config dictionary is valid
                self['variables'][k] = v
            else:
                raise TypeError(f'cannot interpret variable {k} with value {v}')

            # Process data types
            if 'dtype' not in self['variables'][k].keys():
                self['variables'][k]['dtype'] = 'float64'

            # Add to corresponding variables 'output', 'independent' or 'input'
            kind = self['variables'][k]['kind'].lower()
            kind = kind if kind in ('output', 'independent') else 'input'
            if self['variables'][k].get('range') is not None:
                self[kind][k] = self['variables'][k]

        # Fill halton variables with single dimensions of n-D halton
        if halton_dim:
            halton = variable_kinds.halton(size=(self['ntrain'], len(halton_dim)))
            for d, (k, entries) in enumerate(halton_dim):
                diff = (entries[1] - entries[0]) if entries else 1
                low = entries[0] if entries else 0
                self['variables'][k]['range'] = diff * halton[:, d].reshape(-1, 1) + low
                self['input'][k] = self['variables'][k]

        # Fill range of output vector
        for k, v in self['output'].items():
            if not isinstance(v['range'], dict):
                v['range'] = {d: None for d in v['range']}
            shape = []
            for d in v['range']:
                self['output'][k]['range'][d] = self['variables'][d]['range']
                shape.append(self['variables'][d]['range'].shape[0])
            self['output'][k]['shape'] = tuple(shape)

        # Run configuration
        if 'run' not in self:
            self['run'] = {}
        if isinstance(self['run'], str):
            self['run'] = {'command': self['run']}
        Runner.handle_run_config(self)

        if self.get('fit'):
            Surrogate.handle_config(self['fit'], self)

        # Set missing mandatory dict entries to default
        if not self['files'].get('input'):
            self['files']['input'] = path.join(self['base_dir'], 'input.txt')
        if not self['files'].get('output'):
            self['files']['output'] = path.join(self['base_dir'], 'output.txt')

        # Set absolute paths
        self['files']['input'] = path.join(self['base_dir'], self['files']['input'])
        self['files']['output'] = path.join(self['base_dir'], self['files']['output'])
        if self['fit'].get('load'):
            self['fit']['load'] = path.join(self['base_dir'], self['fit']['load'])
        if self['fit'].get('save'):
            self['fit']['save'] = path.join(self['base_dir'], self['fit']['save'])
        return self

    def _remove_nones(self,config=None):
        if config==None: config=self.__dict__
        for key in list(config):
            if type(config[key]) is dict:
                self._remove_nones(config[key])
            #elif (type(config[key]) is not list) and (config[key] is None):
            else:
                if config[key] is None:
                    del config[key]



"""Backends for uncertainty quantification.

TODO: Refactor UQ part
"""

class ChaosPy:
    import chaospy as cp

    def __init__(self, order, sparse=False):
        self.Normal = self.cp.Normal
        self.Uniform = self.cp.Uniform
        self.order = order
        self.sparse = sparse

    def get_eval_points(self, params):
        distribution = self.cp.J(*params.values())
        nodes, _ = self.cp.generate_quadrature(self.order+1, distribution,
                                               rule='G', sparse=self.sparse)
        return nodes



from . import runner
from . import worker

from . import default
from . import zeromq
from . import slurm

from .runner import Runner, RunnerInterface
from .worker import Worker, Interface, Preprocessor, Postprocessor



import numpy as np
import matplotlib.pyplot as plt
from profit.sur.backend.gp_functions_old import invert, nll, predict_f, \
    get_marginal_variance_BBQ, plot_searching_phase
from profit.sur.backend.gp_functions_old import k as kern_sqexp
from scipy.optimize import minimize


def f(x): return x*np.cos(10*x)


# Custom function to build GP matrix
def build_K(xa, xb, hyp, K):
    for i in np.arange(len(xa)):
        for j in np.arange(len(xb)):
            K[i, j] = kern_sqexp(xa[i], xb[j], hyp[0])


def nll_transform(log10hyp):
    hyp = 10**log10hyp
    return nll(hyp, xtrain, ytrain, 1, build_K=build_K)


def sigmoid(x):
    return 1.0/(1.0 + np.exp(-x))


def prior(hyp):
    return sigmoid(hyp[0]-6)*sigmoid(hyp[-1]-6)


noise_train = 0.01

total_train = 15
dimension = 1
hyp = [0.5, 1e-2]
ntest = 100
xtest = np.linspace(0, 1, ntest)
start_point = ntest/2 # because it's always better to begin at the center
next_candidate = start_point

ftest = f(xtest)
list_of_observation = []

n_candidate = 20 # = ntest

for ntrain in range(1, total_train + 1):

    print("\n\nITERATION ", ntrain)

    list_of_observation.append(xtest[int(next_candidate)])

    xtrain = np.array(list_of_observation).reshape((ntrain, dimension))


    ftrain = f(xtrain)
    np.random.seed(0)
    ytrain = ftrain + noise_train*np.random.randn(ntrain, 1)

    K = np.empty((ntrain, ntrain))   # train-train
    build_K(xtrain, xtrain, hyp, K)  # writes inside K
    Ky = K + hyp[-1]*np.eye(ntrain)
    Kyinv = invert(Ky, 4, 1e-6)       # using gp_functions.invert

    Ks = np.empty((ntrain, ntest))  # train-test
    Kss = np.empty((ntest, ntest))  # test-test
    build_K(xtrain, xtest, hyp, Ks)
    build_K(xtest, xtest, hyp, Kss)

    fmean = Ks.T.dot(Kyinv.dot(ytrain)) # predictive mean

    Ef, varf = predict_f(hyp, xtrain.reshape(-1, 1),
                  ytrain.reshape(-1, 1), xtest.reshape(-1, 1), neig=8)# posterior
    # Estimation and variance
    varf = np.diag(varf)

    ############ OPTIMIZER #########
    res = minimize(nll_transform, np.array([np.log10(hyp[0]), np.log10(hyp[1])]), method='BFGS')
    new_hyp = [10**res.x[0], 10**res.x[1]]
    hess_inv = res.hess_inv
    ################################

    ########### MARGINAL VARIANCE #############
    marginal_variance = get_marginal_variance_BBQ(hess_inv, new_hyp, ntrain, ntest, xtrain, xtest,
                                      Kyinv, ytrain, varf,False)
    ###########################################
    varf= varf.reshape((ntest, 1)) # -> to avoid broadcasting

    ######################### PLOT ########################

    plt.plot(xtrain, ytrain, 'kx')
    plt.plot(xtest, ftest, 'm-')
    plt.plot(xtest, fmean, 'r--')
    axes = plt.gca()
    axes.set_ylim([-1.5, 1])
    plt.title('Gaussian Process with '+ str(ntrain) + ' observation(s) hyp = [0.1, 1e-4]')
    plt.legend(('training', 'reference', 'prediction'))
    if ntrain == 10:
        plt.savefig('Active Gaussian Process with '+ str(ntrain) + ' observation(s)')



    plt.fill_between(xtest, # x
                     (fmean.flatten() + 2 * np.sqrt(np.abs(varf).flatten())), # y1
                     (fmean.flatten() - 2 * np.sqrt(np.abs(varf).flatten())), facecolor='blue', alpha=0.4) # y2

    plt.fill_between(xtest, # x
                     (fmean.flatten() + 2 * np.sqrt(marginal_variance.flatten())), # y1
                     (fmean.flatten() - 2 * np.sqrt(marginal_variance.flatten())), facecolor='yellow', alpha=0.4) # y2

    ################ FIND NEXT POINT ################

    scores = marginal_variance + varf
    print(np.c_[marginal_variance, varf, scores])
    next_candidate = np.argmax(scores)

    plot_searching_phase(scores, xtest, next_candidate, ntrain)
    #plt.show()

    #################################################

    hyp = new_hyp
    print("next candidate ->\nx = ", xtest[next_candidate])

plt.show()



import numpy as np
from pyccel.decorators import types, python


@types('real[:]', 'real[:]', 'real')
def kern_sqexp(x0, x1, h): #h: length scale
    """Squared exponential kernel"""
    ret = np.real(np.exp(-0.5*np.sum(((x1 - x0)/h)**2)))
    if ret<1e-16:
        return 0.0
    return ret


@types('real[:]', 'real[:]', 'real[:]')
def kern_sqexp_multiscale(x0, x1, h):
    """Squared exponential kernel with different scale in each direction"""
    return np.real(np.exp(-0.5*np.sum(((x1 - x0)/h)**2)))


@types('real[:]', 'real[:]', 'real')
def kern_wendland4(x0, x1, h):
    """Wendland kernel, positive definite for dimension <= 3"""
    r = np.sqrt(np.sum(((x1 - x0)/h)**2))
    if r < 1.0:
        ret = (1.0 - r)**4*(1.0 + 4.0*r)
    else:
        ret = 0.0
    return ret


@types('real[:]', 'real[:]', 'real[:]')
def kern_wendland4_multiscale(x0, x1, h):
    """Wendland kernel, positive definite for dimension <= 3, 
       different scale in each direction"""
    r = np.real(np.sqrt(np.sum(((x1 - x0)/h)**2)))
    if r < 1.0:
        ret = np.abs((1.0 - r)**4*(1.0 + 4.0*r))
    else:
        ret = 0.0
    return ret


@python
@types('real[:]', 'real[:]', 'real')
def kern_wendland4_product(x0, x1, h):
    """Wendland product kernel"""
    raise NotImplementedError(
        "Upstream bug https://github.com/pyccel/pyccel/issues/245")
    dx = np.zeros_like(x0)
    dx[:] = np.abs(x1 - x0)/h
    outside = False
    nx = len(x0)
    for kx in range(nx):
        dxk = dx[kx]
        if dxk > 1.0:
            outside = True
            break
    if outside:
        ret = 0.0
    else:
        ret = np.abs(np.prod((1.0 - dx)**4*(1.0 + 4.0*dx)))
    return ret


@python
@types('real[:]', 'real[:]', 'real[:]')
def kern_wendland4_product_multiscale(x0, x1, h):
    """Wendland product kernel, different scale in each direction"""
    raise NotImplementedError(
        "Upstream bug https://github.com/pyccel/pyccel/issues/245")
    dx = np.zeros_like(x0)
    dx[:] = np.abs(x1 - x0)/h
    outside = False
    nx = len(dx)
    for kx in range(nx):
        dxk = dx[kx]
        if dxk > 1.0:
            outside = True
            break
    if outside:
        ret = 0.0
    else:
        ret = np.abs(np.prod((1.0 - dx**4)*(1.0 + 4.0*dx)))
    return ret


@types('real[:,:]', 'real[:,:]', 'real[:]', 'real[:,:]')
def gp_matrix(x0, x1, a, K):
    """Constructs GP covariance matrix between two point tuples x0 and x1"""
    n0 = len(x0)
    n1 = len(x1)
    for k0 in range(n0):
        for k1 in range(n1):
            K[k0, k1] = a[1]*kern_sqexp(x0[k0, :], x1[k1, :], a[0])


@types('real[:,:]', 'real[:,:]', 'real', 'real[:,:]')
def gp_matrix_wend(x0, x1, l, K):
    n0 = len(x0)
    n1 = len(x1)
    for k0 in range(n0):
        for k1 in range(n1):
            K[k0, k1] = kern_wendland4(x0[k0, :], x1[k1, :], l)


@python
def gp_matrix_gen(x0, x1, hyp, kern, K):
    """Constructs GP covariance matrix between two point tuples x0 and x1
       with a general kernel kern with hyperparameters hyp
    """
    n0 = len(x0)
    n1 = len(x1)
    for k0 in range(n0):
        for k1 in range(n1):
            K[k0, k1] = kern(x0[k0, :], x1[k1, :], hyp)



"""
Module which includes kernels for the Custom surrogate.
"""

import numpy as np


def RBF(X, Y, length_scale=1, sigma_f=1, sigma_n=0, eval_gradient=False):
    r"""Squared exponential kernel, also call Radial-Basis-Function kernel (RBF).

    The RBF kernel is one of the most common kernels used and is especially suited for smooth functions.

    $$
    \begin{equation}
    K(X, Y) = \sigma_f^2 \exp(-\frac{1}{2} \frac{\lvert X-Y \rvert}{l^2}) + \sigma_n^2~\mathbf{1}
    \end{equation}
    $$

    Parameters:
        X (ndarray): Input points.
        Y (ndarray): Other input points.
        length_scale (float/ndarray): Length scale $l$ of the kernel function.
        sigma_f (float): Scale $\sigma_f$.
        sigma_n (float): Additive noise $\sigma_n$.
        eval_gradient (bool): Indicates if the gradient with respect to the hyperparameters $l$, $\sigma_f$ and
            $\sigma_n$ should be returned.

    Returns:
        tuple/ndarray: A tuple containing:
            - K (ndarray): Kernel matrix of size (X.shape[0], Y.shape[0]).
            - dK (ndarray): Derivative of the kernel w.r.t. the hyperparameters $l$, $\sigma_f$ and $\sigma_n$.
            If eval_gradient is False, only K is returned.
    """

    x1 = X / length_scale
    x2 = Y / length_scale
    dx = x1[:, np.newaxis, :] - x2[np.newaxis, :, :]  # Distance between points in all dimensions.
    dx2 = np.linalg.norm(dx, axis=-1) ** 2
    K = sigma_f ** 2 * np.exp(-0.5 * dx2) + sigma_n ** 2 * np.eye(*dx2.shape)
    if eval_gradient:
        len_length_scale = len(length_scale) if hasattr(length_scale, '__iter__') else 1  # Check for isotropic kernel.
        dK = np.empty((K.shape[0], K.shape[1], len_length_scale + 2))
        K_pure = K - sigma_n ** 2 * np.eye(*K.shape)  # Kernel without noise sigma_n
        if len_length_scale == 1:
            dK[..., 0] = dx2 / length_scale * K_pure  # dK/dl
        else:
            dK[..., :-2] = dx ** 2 / length_scale * K_pure[..., np.newaxis]  # dK/dl for n-D length_scale
        dK[..., -2] = 2 / sigma_f * K_pure  # dK/dsigma_f
        dK[..., -1] = 2 * sigma_n * np.eye(*K_pure.shape)  # dK/dsigma_n
        return K, dK
    return K


def LinearEmbedding(X, Y, R, sigma_f=1e-6, sigma_n=0, eval_gradient=False):
    r"""The linear embedding kernel according to Garnett (2014) is a generalisation of the RBF kernel.

    The RBF kernel with different length scales in each dimension (ARD kernel) is a special case where only
    the diagonal elements of the matrix $R$ are non-zero and represent inverse length scales.
    The ARD kernel is suited to detect relevant dimensions of the input data. In contrast,
    the linear embedding kernel can also detect relevant linear combinations of dimensions, e.g. if a function only
    varies in the direction $x1 + x2$. Thus, the linear embedding kernel can be used to find a lower dimensional
    representation of the data in aribitrary directions.

    $$
    \begin{equation}
    K(X, Y) = \sigma_f^2 \exp(-\frac{1}{2} (X-Y)R^T R(X-Y)^T) + \sigma_n^2~\mathbf{1}
    \end{equation}
    $$

    Here, we use the convention that the data $X$ and $Y$ are of shape (n, D) and therefore R has to be of shape
    (d, D) where D > d to find a lower dimensional representation.

    Parameters:
        X (ndarray): Input points of shape (n, D)
        Y (ndarray): Other input ponits.
        R (ndarray): Matrix or flattened array of hyperparameters. It is automatically reshaped to fit the input data.
            Every matrix element represents a hyperparameter.
        sigma_f (float): Scale $\sigma_f$.
        sigma_n (float): Additive noise $\sigma_n$.
        eval_gradient (bool): Indicates if the gradient with respect to the hyperparameters $R_{ij}$, $\sigma_f$ and
            $\sigma_n$ should be returned.

   Returns:
        tuple/ndarray: A tuple containing:
            - K (ndarray): Kernel matrix of size (X.shape[0], Y.shape[0]).
            - dK (ndarray): Derivative of the kernel w.r.t. the hyperparameters $R_{ij}$, $\sigma_f$ and $\sigma_n$.
            If eval_gradient is False, only K is returned.
    """

    X = np.atleast_2d(X)
    # X @ R.T is of shape (n x D) @ (D x d)  -> second dimension of R must be D, so first must be R.size / D
    R = R.reshape(R.size // X.shape[-1], X.shape[-1])
    X1 = X @ R.T
    X2 = X @ R.T if Y is None else Y @ R.T
    dX = X1[:, np.newaxis, :] - X2[np.newaxis, :, :]
    dX2 = np.linalg.norm(dX, axis=-1) ** 2
    K = sigma_f ** 2 * np.exp(-0.5 * dX2) + sigma_n ** 2 * np.eye(*dX2.shape)
    if eval_gradient:
        # TODO: Check gradient dK/dR. For a 2x2 R it should be (4,) dimensional.
        dK = np.empty((K.shape[0], K.shape[1], R.size + 2))
        K_pure = K - sigma_n ** 2 * np.eye(*K.shape)  # Kernel without noise sigma_n
        dK[..., :-2] = np.einsum('ijk,kl', dX ** 2, R) * K_pure[..., np.newaxis]  # dK/dR_ij
        dK[..., -2] = 2 / sigma_f * K_pure  # dK/dsigma_f
        dK[..., -1] = 2 * sigma_n * np.eye(*K_pure.shape)  # dK/dsigma_n
        return K, dK
    return K



#%%
from sympy import *
from sympy.abc import i, j, k, l, N
from sympy.utilities.codegen import codegen

n, m = symbols('n m', integer=True)
M = MatrixSymbol("M", n, m)
b = MatrixSymbol("b", m, 1)

expr = Sum(b[k, 0]*b[k, 0], (k, 0, m-1))
dexpr = diff(expr, b[j, 0])
print(dexpr)

funlist = [
    ('expr', expr),
    ('dexpr', dexpr)
]
[(name, code), (h_name, header)] = codegen(
    funlist, 'F95', 'kernels',  argument_sequence=(j, k, m, b), header=False, empty=False)

print(code)

# %%



from .halton import halton as _halton_base
from .util import check_ndim
import numpy as np


def halton(size=(1, 1)):
    if isinstance(size, (tuple, list, np.ndarray)):
        return _halton_base(*size)
    else:
        return check_ndim(_halton_base(size, 1))


def uniform(start=0, end=1, size=None):
    return check_ndim(start + np.random.random(size) * (end - start))


def loguniform(start=1e-6, end=1, size=None):
    return check_ndim(start * np.exp((np.log(end) - np.log(start)) * np.random.random(size)))


def normal(mu=0, std=1, size=None):
    return check_ndim(mu + std * np.random.randn(size))


def linear(start=0, end=1, step=1, size=None):
    return check_ndim(np.arange(start, end, step))


def independent(start=0, end=1, step=1, size=None):
    return linear(start, end, step)


def activelearning(size=None):
    return check_ndim(np.full(size, np.nan))


def constant(value=0, size=None):
    return check_ndim(np.full(size, value))



"""

Explicit imports:
    profit.fit
    profit.run
    profit.sur
    profit.ui
    profit.uq
"""

from . import main
from . import pre
from . import post
from . import config

from . import util

# Retrieving package version at runtime
# https://pypi.org/project/setuptools-scm/
try:
    from importlib.metadata import version, PackageNotFoundError
except ModuleNotFoundError:  # required for python3.7
    from importlib_metadata import version, PackageNotFoundError

try:
    __version__ = version("profit")
except PackageNotFoundError:
    # package is not installed
    pass

del version, PackageNotFoundError  # clean up namespace



import numpy as np

from profit.sur.linear_reduction import KarhunenLoeve, PCA

np.random.seed(42)

def f(x, t):
    return np.cos(x*t)

t = np.linspace(0, 20, 100)

ntrain = 50

xtrain = np.random.rand(ntrain)
ytrain = []
for x in xtrain:
    ytrain.append(f(x, t))
ytrain = np.array(ytrain)

ymean = np.mean(ytrain, 0)
dy = ytrain - ymean

tol = 1e-3

xtest = 0.7
ytest = f(xtest, t)


def test_karhunen_loeve():
    kl_model = KarhunenLoeve(ytrain, tol)
    z = kl_model.project(ytest)
    ylift = kl_model.lift(z)

    w, Q = np.linalg.eigh(dy @ dy.T)
    assert(np.allclose(kl_model.w, w[w>tol]))
    assert(np.allclose(kl_model.Q, Q[:, w>tol]))
    assert(np.max(np.abs(ytest - ylift)) < tol)


def test_karhunen_loeve_vec():
    kl_model = KarhunenLoeve(ytrain, tol)
    xtest = np.array([0.7, 0.8, 0.9])
    ytest = np.array([f(x, t) for x in xtest])
    z = kl_model.project(ytest)
    ylift = kl_model.lift(z)
    assert(np.max(np.abs(ytest - ylift)) < tol)


def test_PCA():
    pca_model = PCA(ytrain, tol)
    z = pca_model.project(ytest)
    ylift = pca_model.lift(z)

    w, Q = np.linalg.eigh(dy.T @ dy)
    assert(np.allclose(pca_model.w, w[w>tol]))
    assert(np.allclose(pca_model.Q, Q[:, w>tol]))
    assert(np.max(np.abs(ytest - ylift)) < tol)


def test_PCA_vec():
    pca_model = PCA(ytrain, tol)
    xtest = np.array([0.7, 0.8, 0.9])
    ytest = np.array([f(x, t) for x in xtest])
    z = pca_model.project(ytest)
    ylift = pca_model.lift(z)
    assert(np.max(np.abs(ytest - ylift)) < tol)



#%%
import numpy as np
import GPy
from GPy.kern import Prod

k0 = GPy.kern.RBF(1, active_dims=0)  # SqExp in first dimension
k1 = GPy.kern.RBF(1, active_dims=1)  # SqExp in second dimension

k01 = GPy.kern.RBF(2)  # SqExp in 2D for comparison
kprod = Prod((k0, k1))

x0train = np.array([0.0, 1.0, 0.0]).reshape(-1,1)
x1train = np.array([0.0, 0.0, 1.0]).reshape(-1,1)
xtrain = np.hstack((x0train, x1train))

print('Prod K = ')
print(kprod.K(xtrain, xtrain))
print()

print('Reference K = ')
print(k01.K(xtrain, xtrain))
print()

# %%



# %% See
# https://pytorch.org/tutorials/beginner/blitz/neural_networks_tutorial.html
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from profit.sur.ann import Autoencoder


class LinearAutoencoder(nn.Module):
    """Linear autoencoder as 'rich man's PCA'.

    Weights of the encoder are tied to the decoder
    to have the symmetric structure x = W.T @ W @ x .
    """
    def __init__(self, D, d):
        super(LinearAutoencoder, self).__init__()
        self.W = nn.Parameter(torch.randn((d, D)) / np.sqrt(d * D))
        #self.W = nn.Parameter(torch.eye(D))

    def forward(self, x):
        x = F.linear(x, self.W)      # encoder
        x = F.linear(x, self.W.t())  # decoder
        return x


D = 4
d = 2

Wlift = torch.Tensor([[1.0, 0], [0, 1.0], [0.0, 0.2], [0.3, 0.0]])

x = Wlift @ torch.rand(d)

ae = Autoencoder(D, d)
print(list(ae.parameters()))
y = ae.forward(x)
print(x)
print(y)
# %%
ntrain = 10
input = (Wlift @ (torch.rand(d, ntrain) - 0.5)).T
output = ae(input)
target = input
criterion = nn.MSELoss()
loss = criterion(output, target)

print(loss)
#%% PyTorch custom plain gradient descent: quite slow
# learning_rate = 0.1
# for k in range(100):
#     ae.zero_grad()
#     output = ae(input)
#     loss = criterion(output, target)
#     loss.backward()
#     for f in ae.parameters():
#         f.data.sub_(f.grad.data * learning_rate)
#     print(loss.data)

#%% PyTorch SGD variant (Adam)
optimizer = optim.Adam(ae.parameters(), lr=0.01)

for k in range(1000):
    optimizer.zero_grad()
    output = ae(input)
    loss = criterion(output, target)
    loss.backward()
    optimizer.step()
    print(loss.data)
#%%
x = Wlift @ (torch.rand(d) - 0.5)
print(x.data)
print(ae(x).data)
#%% PyTorch Quasi-Newton (LBFGS)

# # create your optimizer
# optimizer = optim.LBFGS(ae.parameters())

# for inp in input:
#     def closure():
#         if torch.is_grad_enabled():
#             optimizer.zero_grad()
#         output = ae(inp)
#         loss = criterion(output, inp)
#         if loss.requires_grad:
#             loss.backward()
#         return loss
#     optimizer.step(closure)

# output = ae(input)
# print(criterion(output, input))

#%% PyTorch SGD variant (Adam)


#%% SciPy Quasi-Newton (Powell)
# from scipy.optimize import minimize

# def target(w):
#     with torch.no_grad():
#         ae.W.copy_(torch.Tensor(w.reshape([d,D])))
#     if ae.W.grad is not None:
#         ae.W.grad.zero_()
#     output = ae(input)
#     loss = criterion(output, input)
#     if loss.requires_grad:
#         loss.backward()
#     with torch.no_grad():
#         print(loss.data)

#     return loss.detach().numpy(), ae.W.grad.detach().flatten().numpy()

# target(ae.W)
# res = minimize(target, ae.W.detach().flatten(), method='Powell', jac=True)
# print(res.fun)
# with torch.no_grad():
#     ae.W.copy_(torch.Tensor(res.x.reshape([d,D])))
#
# print()
# print("Weights: ")
# print(ae.W @ ae.W.T)
# print(ae.W.T @ ae.W)
# %%

# %%



"""Interfaces for uncertainty quantification.

TODO: Refactor UQ part
"""

import os
from collections import OrderedDict
from profit.config import Config

def read_params(filename):
    from numpy import genfromtxt
    global params
    data = genfromtxt(os.path.join(config.base_dir, filename),
                         dtype=None, encoding=None)
    for param in data:
        if(param[1] == 'Normal'):
            params[param[0]] = Normal(param[2], param[3])

class UQ:
    def __init__(self, config=None, yaml=None):
        from chaospy import (generate_quadrature, orth_ttr, fit_quadrature, E, Std,
            descriptives)
        self.params = OrderedDict()
        self.backend = None
        self.param_files = None

        if yaml:
            print('  load configuration from %s'%yaml)
            config = Config.from_file(yaml)

        if config:
            if (config['uq']['backend'] == 'ChaosPy'):
              self.backend = ChaosPy(config['uq']['order'])
              # TODO: extend

            self.Normal = self.backend.Normal
            self.Uniform = self.backend.Uniform

            params = config['uq']['params']
            for pkey in params:
              if params[pkey]['dist'] == 'Uniform':
                self.params[pkey] = self.Uniform(params[pkey]['min'],
                                                 params[pkey]['max'])
            if 'param_files' in config['uq']:
              self.param_files = config['uq']['param_files']

        self.template_dir = 'template/'
        self.run_dir = 'run/'

    def write_config(self, filename='profit.yaml'):
        '''
        write UQ-configuration to yaml file.
        The SLURM configuration is so far not dumped yet'
        '''
        config = self.get_config()
        config.write_yaml(filename)

    def get_config(self):
        config = Config()
        configuq = config['uq']
        if isinstance(self.backend,ChaosPy):
          configuq['backend'] = 'ChaosPy'
          configuq['order'] = self.backend.order
          configuq['sparse'] = self.backend.sparse

        configuq['params'] = OrderedDict()
        for param in self.params:
          p = self.params[param]
          if isinstance(p,self.backend.Uniform):
            configuq['params'][param]={'dist':'Uniform','min':float(p.range()[0]),'max':float(p.range()[1])}
          elif isinstance(p,self.backend.Normal):
            configuq['params'][param]={'dist':'Normal'}

        configuq['param_files']=self.param_files
        config['run_dir']=self.run_dir
        config['template_dir']=self.template_dir

        return config

    def write_input(self, run_dir='run/'):
        '''
        write input.txt with parameter combinations to
        directory "run_dir"
        '''
        from numpy import savetxt
        self.eval_points = self.backend.get_eval_points(self.params)
        savetxt(os.path.join(run_dir, 'input.txt'),
                self.eval_points.T, header=' '.join(self.params.keys()))


    def get_eval_points(self):
        """Returns N evaluation points in P parameters as array of shape (P,N)"""
        return self.backend.get_eval_points(self.params)



"""
Created: Fri Jul 26 09:57:16 2019
@author: Christopher Albert <albert@alumni.tugraz.at>
"""

import time

from chaospy import Normal, J, generate_quadrature

params = [Normal(mu=0.999,  sigma=0.0052), 
          Normal(mu=27.1,   sigma=17.0), 
          Normal(mu=0.318,  sigma=0.1),
          Normal(mu=0.015,  sigma=0.0087),
          Normal(mu=0.0817, sigma=0.0077),
          Normal(mu=1.309,  sigma=0.086),
          Normal(mu=2.19,   sigma=0.22)]

dist = J(*params)
#%%
t = time.time()
nodes, weights = generate_quadrature(4, dist, rule='G', sparse=True)
print('time elapsed: {} s'.format(time.time() - t))



ntrain = 2
variables = {'u': 'Halton()',
             'v': 'Uniform(0.55, 0.6)',
             'w': 'ActiveLearning()',
             'r': 'Independent(0, 1, 0.1)',
             'f': {'kind': 'Output', 'depend': ('r',), 'range': 'r'},
             'g': 'Output(r)'}

files = {'input': 'input.txt',
         'output': 'output.txt'}

run = {'pre': {'class': 'template', 'param_files': ['mockup.in']},
       'post': {'class': 'numpytxt', 'path': 'mockup.out'},
       'command': 'python3 mockup.py',
       'clean': False}



"""Mockup domain code."""
import numpy as np


def rosenbrock(x, y, a, b):
    return (a - x)**2 + b * (y - x**2)**2


def f(r, u, v):
    return rosenbrock((r - 0.5) + u - 5, 1 + 3 * (v - 0.6), a=1, b=3), r


params = np.loadtxt('mockup.in')
r = np.arange(0, 1, 0.1)
result = f(r, params[0], params[1])
print(result)
np.savetxt('mockup.out', np.concatenate([result[0], result[1]]))



"""Utility functions.

This file contains functions for loading and saving,
string processing as well as random sampling.
"""
from os import path
from typing import Union
from collections.abc import MutableMapping, Mapping
import numpy as np

def save(filename, data, header=None, fmt=None):

    if filename.endswith('.txt'):
        save_txt(filename, data, header, fmt)
    elif filename.endswith('.hdf5'):
        save_hdf(filename, data)


def load(filename, as_type='dtype'):
    if filename.endswith(('.txt', '.in', '.out')):
        return load_txt(filename, names=True if as_type == 'dtype' else None)
    elif filename.endswith('.hdf5'):
        return load_hdf(filename, as_type)


def load_txt(filename, names=True):
    from numpy import genfromtxt
    return check_ndim(genfromtxt(filename, names=names))


def save_txt(filename, data, header=None, fmt=None):
    from numpy import hstack, savetxt
    if not header:
        header = ' '.join(data.dtype.names)
    data = hstack([data[key] for key in data.dtype.names])
    if fmt:
        savetxt(filename, data, header=header, fmt=fmt)
    else:
        savetxt(filename, data, header=header)


def save_hdf(filename, data):
    """ Save data to a hdf5 file.
    HDF5 keys according to either numpy dtypes, dict keys or 'data' if none of the above is true. """
    from h5py import File
    with File(filename, 'w') as h5f:
        if hasattr(data, 'dtype'):
            for key in data.dtype.names:
                h5f[key] = data[key]
        elif isinstance(data, dict):
            def recursive_dict2hdf(_path, _dict):
                for key, value in _dict.items():
                    if isinstance(value, dict):
                        recursive_dict2hdf(_path + str(key) + '/', value)
                    else:
                        h5f[_path + key] = value
            recursive_dict2hdf('', data)
        else:
            h5f['data'] = data


def load_hdf(filename, astype='dtype'):
    from h5py import File
    from numpy import array
    with File(filename, 'r') as h5f:
        if astype == 'dtype':
            return hdf2numpy(h5f)
        if astype == 'dict':
            return hdf2dict(h5f)
        else:
            return array(h5f['data'])


def txt2hdf(txtfile, hdffile):
    save_hdf(hdffile, load_txt(txtfile))


def hdf2txt(txtfile, hdffile):
    save_txt(txtfile, load_hdf(hdffile))


def hdf2numpy(dataset):
    from numpy import zeros_like, array
    dtypes = [(key, float) for key in list(dataset.keys())]
    data = zeros_like(dataset[dtypes[0][0]], dtype=dtypes)
    for key in data.dtype.names:
        data[key] = array(dataset[key])
    return check_ndim(data)


def hdf2dict(dataset):
    from numpy import array, ndarray, atleast_1d
    from h5py import Dataset
    load_dict = {}

    def recursive_hdf2dict(_data, _dict):
        for key in _data.keys():
            if isinstance(_data[key], Dataset):
                val = _data[key][()]
                if isinstance(val, bytes):  # quick fix for new h5py version, which stores strings as bytes
                    val = val.decode('utf-8')
                _dict[key] = atleast_1d(array(val)) if isinstance(val, ndarray) else val
            else:
                _dict[key] = {}
                recursive_hdf2dict(_data[key], _dict[key])
    recursive_hdf2dict(dataset, load_dict)
    return load_dict


def safe_path_to_file(arg, default, valid_extensions=('.yaml', '.py')):

    if path.isfile(arg):
        if arg.endswith(valid_extensions):
            return path.abspath(arg)
        else:
            raise TypeError("Unsupported file extension. \n"
                            "Valid file extensions: {}".format(arg, valid_extensions))
    elif path.isdir(arg):
        return path.join(arg, default)
    else:
        raise FileNotFoundError("Directory or file ({}) not found.".format(arg))


def safe_str(string, lower=True, replace=('-', '_', ' '), replace_with=''):
    if lower:
        string = string.lower()
    if replace:
        for r in replace:
            string = string.replace(r, replace_with)
    return string


def get_class_methods(cls):
    return [func for func in dir(cls) if callable(getattr(cls, func)) and not func.startswith("_")]


def get_class_attribs(self):
    return [attr for attr in dir(self) if not callable(getattr(self, attr)) and not attr.startswith('_')]


def quasirand(ndim=1, npoint=1):
    from .halton import halton
    return halton(npoint, ndim)


def check_ndim(arr):
    return arr if arr.ndim > 1 else arr.reshape(-1, 1)


class SafeDict(dict):
    def __init__(self, obj, pre='{', post='}'):
        self.pre = pre
        self.post = post
        super().__init__(obj)

    @classmethod
    def from_params(cls, params, **kwargs):
        return cls(params2map(params), **kwargs)

    def __missing__(self, key):
        return self.pre + key + self.post


def params2map(params: Union[None, MutableMapping, np.ndarray, np.void]):
    if params is None:
        return {}
    if isinstance(params, MutableMapping):
        return params
    try:
        return {key: params[key] for key in params.dtype.names}
    except AttributeError:
        pass
    raise TypeError('params are not a Mapping')


def spread_struct_horizontal(struct_array: np.ndarray, variable_config: Mapping):
    dtype = []
    columns = {}
    # prepare dtype
    for variable in struct_array.dtype.names:
        spec = variable_config[variable]
        if len(spec['shape']) == 0:
            dtype.append((variable, spec['dtype']))
            columns[variable] = [variable]
        else:
            ranges = []
            columns[variable] = []
            for dep in spec['depend']:
                ranges.append(spec['range'][dep])
            meshes = [m.flatten() for m in np.meshgrid(*ranges)]
            for i in range(meshes[0].size):
                name = variable + '(' + ', '.join([f'{m[i]}' for m in meshes]) + ')'
                dtype.append((name, spec['dtype']))
                columns[variable].append(name)
    # fill data
    output = np.zeros(struct_array.shape, dtype=dtype)
    for variable, spec in variable_config.items():
        if len(spec['shape']) == 0:
            output[variable] = struct_array[variable]
        else:
            for i in range(struct_array.size):
                output[columns[variable]][i] = tuple(struct_array[variable][i])
    return output


# ToDo: spread struct vertical
#  -> independent variables get new columns
#  -> 1 original row gets spread across several (with duplicate entries)


def flatten_struct(struct_array: np.ndarray):
    # per default vector entries are spread across several columns
    if not struct_array.size:
        return np.array([[]])
    return np.vstack([np.hstack([row[key].flatten() for key in struct_array.dtype.names]) for row in struct_array])


def load_includes(paths):
    """ load python modules from the specified paths """
    import os
    import sys
    from importlib.util import spec_from_file_location, module_from_spec
    import logging
    for path in paths:
        name = f"profit_include_{os.path.basename(path).split('.')[0]}"
        if name in sys.modules:  # do not reload modules
            continue
        try:
            spec = spec_from_file_location(name, path)
        except FileNotFoundError:
            logging.getLogger(__name__).error(f'could not find {path} to include')
            continue
        module = module_from_spec(spec)
        sys.modules[name] = module
        spec.loader.exec_module(module)



# -*- coding: utf-8 -*-
"""
Created on Tue Apr 21 16:47:00 2020

@author: Katharina Rath
"""

import numpy as np
from numpy import exp, sin
import matplotlib.pyplot as plt
import scipy.integrate as spint
from scipy.optimize import fsolve, fmin, minimize, root, newton
from scipy.linalg import solve_triangular
from scipy.sparse.linalg import eigsh
#from sympy import *
#from sympy.utilities.autowrap import ufuncify
import scipy

from scipy.integrate import solve_ivp
from param import nm, U0, dtsymp


# Load precompiled ufuncified stuff from sympy
#tmp = 'tmp'

#import sys
#if not tmp in sys.path:
#    sys.path.append(tmp)

from kernels import *


def f_kern(x, y, x0, y0, l):
    return kern_num(x,y,x0,y0,l[0], l[1])

def dkdx(x, y, x0, y0, l):
    return dkdx_num(x,y,x0,y0,l[0], l[1])

def dkdy(x, y, x0, y0, l):
    return dkdy_num(x,y,x0,y0,l[0], l[1])

def dkdx0(x, y, x0, y0, l):
    return dkdx0_num(x,y,x0,y0,l[0], l[1])

def dkdy0(x, y, x0, y0, l):
    return dkdy0_num(x,y,x0,y0,l[0], l[1])

def d2kdxdx0(x, y, x0, y0, l):
    return d2kdxdx0_num(x,y,x0,y0,l[0], l[1])

def d2kdydy0(x, y, x0, y0, l):
    return d2kdydy0_num(x,y,x0,y0,l[0], l[1])

def d2kdxdy0(x, y, x0, y0, l):
    return d2kdxdy0_num(x,y,x0,y0,l[0], l[1])

def d2kdydx0(x, y, x0, y0, l):
    return d2kdxdy0(x, y, x0, y0, l)

def d3kdxdx0dy0(x, y, x0, y0, l):
    return d3kdxdx0dy0_num(x,y,x0,y0,l[0], l[1])

def d3kdydy0dy0(x, y, x0, y0, l):
    return d3kdydy0dy0_num(x,y,x0,y0,l[0], l[1])

def d3kdxdy0dy0(x, y, x0, y0, l):
    return d3kdxdy0dy0_num(x,y,x0,y0,l[0], l[1])

def d3kdydx0dy0(x, y, x0, y0, l):
    return d3kdxdy0dy0(x, y, x0, y0, l)

def dkdlx(x, y, x0, y0, l):
    return dkdlx_num(x, y, x0, y0, l[0], l[1])

def dkdly(x, y, x0, y0, l):
    return dkdly_num(x, y, x0, y0, l[0], l[1])

def d3kdxdx0dlx(x, y, x0, y0, l):
    return d3kdxdx0dlx_num(x,y,x0,y0,l[0], l[1])

def d3kdydy0dlx(x, y, x0, y0, l):
    return d3kdydy0dlx_num(x,y,x0,y0,l[0], l[1])

def d3kdxdy0dlx(x, y, x0, y0, l):
    return d3kdxdy0dlx_num(x,y,x0,y0,l[0], l[1])

def d3kdydx0dlx(x, y, x0, y0, l):
    return d3kdxdy0dlx(x, y, x0, y0, l)

def d3kdxdx0dly(x, y, x0, y0, l):
    return d3kdxdx0dly_num(x,y,x0,y0,l[0], l[1])

def d3kdydy0dly(x, y, x0, y0, l):
    return d3kdydy0dly_num(x,y,x0,y0,l[0], l[1])

def d3kdxdy0dly(x, y, x0, y0, l):
    return d3kdxdy0dly_num(x,y,x0,y0,l[0], l[1])

def d3kdydx0dly(x, y, x0, y0, l):
    return d3kdxdy0dly(x, y, x0, y0, l)

def intode(y, t, dtsymp):
    ys = np.zeros([2, len(t)])
    ys[0,0] = y[0]
    ys[1,0] = y[1]            

    for kt in range(1,len(t)):
            ys[1, kt]=ys[1, kt-1] - dtsymp*(np.sin(ys[0, kt-1] + np.pi))#-0.1*np.sin(ys[0,kt-1]-0.5*t[kt]))
            ys[0, kt]=ys[0, kt-1] + dtsymp*ys[1, kt]
            
    # ys[0,:] = np.mod(ys[0,:], 2*np.pi)
    return ys.T

def build_K(xin, x0in, l, K):
    # set up covariance matrix
    N = K.shape[1]//2
    N0 = K.shape[0]//2
    x0 = x0in[0:N]
    x = xin[0:N]
    y0 = x0in[N:2*N]
    y = xin[N:2*N]
    for k in range(N0):
        for lk in range(N):
            K[k,lk] = d2kdxdx0(
                x0[k], y0[k], x[lk], y[lk], l) 
            K[N0+k,lk] = d2kdxdy0(
                 x0[k], y0[k], x[lk], y[lk], l) 
            K[k,N+lk] = d2kdydx0(
                 x0[k], y0[k], x[lk], y[lk], l) 
            K[N0+k,N+lk] = d2kdydy0(
                x0[k], y0[k], x[lk], y[lk], l) 
            
def buildK(x, y, x0, y0, l, K):
    # set up covariance matrix
    N = K.shape[1]//2
    N0 = K.shape[0]//2
    
    for k in range(N0):
        for lk in range(N):
            K[k,lk] = d2kdxdx0(
                x0[k], y0[k], x[lk], y[lk], l) 
            K[N0+k,lk] = d2kdxdy0(
                 x0[k], y0[k], x[lk], y[lk], l) 
            K[k,N+lk] = d2kdydx0(
                 x0[k], y0[k], x[lk], y[lk], l) 
            K[N0+k,N+lk] = d2kdydy0(
                x0[k], y0[k], x[lk], y[lk], l) 


def build_dK(x, y, x0, y0, l):
    # set up covariance matrix
    N = len(x)
    N0 = len(x0)
    k11 = np.empty((N0, N))
    k12 = np.empty((N0, N))
    k21 = np.empty((N0, N))
    k22 = np.empty((N0, N))

    dK = []
    
    for k in range(N0):
        for lk in range(N):
              k11[k,lk] = d3kdxdx0dlx(
                  x0[k], y0[k], x[lk], y[lk], l) 
              k21[k,lk] = d3kdxdy0dlx(
                  x0[k], y0[k], x[lk], y[lk], l) 
              k12[k,lk] = d3kdydx0dlx(
                  x0[k], y0[k], x[lk], y[lk], l) 
              k22[k,lk] = d3kdydy0dlx(
                  x0[k], y0[k], x[lk], y[lk], l) 
        
    dK.append(np.vstack([
        np.hstack([k11, k12]),
        np.hstack([k21, k22])
    ]))

    for k in range(N0):
        for lk in range(N):
             k11[k,lk] = d3kdxdx0dly(
                 x0[k], y0[k], x[lk], y[lk], l) 
             k21[k,lk] = d3kdxdy0dly(
                 x0[k], y0[k], x[lk], y[lk], l) 
             k12[k,lk] = d3kdydx0dly(
                  x0[k], y0[k], x[lk], y[lk], l) 
             k22[k,lk] = d3kdydy0dly(
                 x0[k], y0[k], x[lk], y[lk], l) 
        
    dK.append(np.vstack([
        np.hstack([k11, k12]),
        np.hstack([k21, k22])
    ]))

    return dK
    

def build_dKdy0(x, y, x0, y0,l):
    # set up covariance matrix
    N = len(x)
    N0 = len(x0)
    k11 = np.empty((N0, N))
    k12 = np.empty((N0, N))
    k21 = np.empty((N0, N))
    k22 = np.empty((N0, N))
    
    for k in range(N0):
        for lk in range(N):
             k11[k,lk] = d3kdxdx0dy0(
                 x0[k], y0[k], x[lk], y[lk], l) 
             k21[k,lk] = d3kdxdy0dy0(
                  x0[k], y0[k], x[lk], y[lk], l ) 
             k12[k,lk] = d3kdydx0dy0(
                  x0[k], y0[k], x[lk], y[lk], l) 
             k22[k,lk] = d3kdydy0dy0(
                 x0[k], y0[k], x[lk], y[lk], l) 
        
    K = np.vstack([
        np.hstack([k11, k12]),
        np.hstack([k21, k22])
    ])
    return K


def nlp_with_grad(hyp, xtrain, ytrain, ztrain, sig_n):
    K = np.empty((2*len(xtrain), 2*len(xtrain)))
    buildK(xtrain, ytrain, xtrain, ytrain, hyp, K)
    Ky = K + sig_n**2*np.diag(np.ones(np.shape(K)[0]))
    Kyinv = np.linalg.inv(Ky)                # invert GP matrix
    alpha = Kyinv.dot(ztrain)
    nlp_val = 0.5*ztrain.T.dot(alpha) + 0.5*np.linalg.slogdet(Ky)[1]

    dK = build_dK(xtrain, ytrain, xtrain, ytrain, hyp)

    nlp_grad = np.array([
        -0.5*alpha.T.dot(dK[0].dot(alpha)) + 0.5*np.trace(Kyinv.dot(dK[0])),
        -0.5*alpha.T.dot(dK[1].dot(alpha)) + 0.5*np.trace(Kyinv.dot(dK[1]))
    ])

    return nlp_val, nlp_grad

def nlp(hyp, xtrain, ytrain, ztrain, sig_n):
    #try:
    #    return nlpgen(hyp, xtrain, ytrain, ztrain, sig_n)
    #except:
    #    pass
    K = np.empty((2*len(xtrain), 2*len(xtrain)))
    buildK(xtrain, ytrain, xtrain, ytrain, hyp, K)
    Ky = K + sig_n**2*np.diag(np.ones(np.shape(K)[0]))
    Kyinv = np.linalg.pinv(Ky)                # invert GP matrix
    return 0.5*ztrain.T.dot(Kyinv.dot(ztrain)) + 0.5*np.linalg.slogdet(Ky)[1]

def gpsolve(Ky, ft):
    L = np.linalg.cholesky(Ky)
    alpha = solve_triangular(
        L.T, solve_triangular(L, ft, lower=True, check_finite=False), 
        lower=False, check_finite=False)

    return L, alpha
 
def nlpgen(hyp, xtrain, ytrain, ztrain, sig_n):
    K = np.empty((2*len(xtrain), 2*len(xtrain)))
    buildK(xtrain, ytrain, xtrain, ytrain, hyp, K)
    # print(np.linalg.eigvals(K))
    Ky = K + sig_n**2*np.diag(np.ones(2*len(xtrain)))  # add noise variance
    L, alpha = gpsolve(Ky, ztrain)
    return 0.5*ztrain.T.dot(alpha) + np.sum(np.log(L.diagonal()))

def energy(x): 
    return x[1]**2/2 + U0*(1 + np.cos(x[0]))

def calc_tau_ex(q,p):
    kappa = energy([q, p])/(2*U0)
    kappinv = 1/kappa
    if kappa < 1:
        tau = 4*scipy.special.ellipk(kappa)
    else:
        tau = 2.0*scipy.special.ellipk(kappinv)/np.sqrt(kappa)
    return tau

def buildKreg(xtrainreg, ytrainreg, lp):

    N = len(xtrainreg)
        # ztrain = np.concatenate((ztrain1, ztrain2))
    Kp = np.empty((N, N))
    
    for k in range(N):
        for lk in range(N):
            Kp[k,lk] = f_kern(
                     xtrainreg[k], ytrainreg[k], xtrainreg[lk], ytrainreg[lk], lp) 
    return Kp

def build_dKreg(xtrainreg, ytrainreg, lp):

    N = len(xtrainreg)
        # ztrain = np.concatenate((ztrain1, ztrain2))
    Kp = np.empty((N, N))
    
    dK = []
    for k in range(N):
        for lk in range(N):
            Kp[k,lk] = dkdlx(
                     xtrainreg[k], ytrainreg[k], xtrainreg[lk], ytrainreg[lk], lp) 

    dK.append(Kp.copy())

    for k in range(N):
        for lk in range(N):
            Kp[k,lk] = dkdly(
                     xtrainreg[k], ytrainreg[k], xtrainreg[lk], ytrainreg[lk], lp) 

    dK.append(Kp.copy())

    return dK

def nlpreg(hyp, xtrain, ytrain, ztrain, sig_n):
    K = buildKreg(xtrain, ytrain, hyp)
    Ky = K + sig_n**2*np.diag(np.ones(np.shape(K)[0]))
    Kyinv = np.linalg.pinv(Ky)                # invert GP matrix
    return 0.5*ztrain.T.dot(Kyinv.dot(ztrain)) + 0.5*np.linalg.slogdet(Ky)[1]

def nlpreg_with_grad(hyp, xtrain, ytrain, ztrain, sig_n):
    K = buildKreg(xtrain, ytrain, hyp)
    dK = build_dKreg(xtrain, ytrain, hyp)
    Ky = K + sig_n**2*np.diag(np.ones(np.shape(K)[0]))
    Kyinv = np.linalg.inv(Ky)                # invert GP matrix
    alpha = Kyinv.dot(ztrain)

    nlp_val = 0.5*ztrain.T.dot(alpha) + 0.5*np.linalg.slogdet(Ky)[1]

    nlp_grad = np.array([
        -0.5*alpha.T.dot(dK[0].dot(alpha)) + 0.5*np.trace(Kyinv.dot(dK[0])),
        -0.5*alpha.T.dot(dK[1].dot(alpha)) + 0.5*np.trace(Kyinv.dot(dK[1]))
    ])

    return nlp_val, nlp_grad

def guessP(x, y, l, xtrainp, ytrainp, ztrainp, Kyinvp):    
    Ntest = 1
    N = len(xtrainp)
    Kstar = np.empty((Ntest, N))
    for k in range(Ntest):
        for lk in range(N):
            Kstar[k,lk] = f_kern(
                 x[k], y[k], xtrainp[lk], ytrainp[lk], l) 
    Ef = Kstar.dot(Kyinvp.dot(ztrainp))
    return Ef

def calcQ(x,y, xtrain, ytrain, l, Kyinv, ztrain):
    # get \Delta q from GP on mixed grid. 
    # temporarily, also \delta p is passed from the function
    Kstar = np.empty((2, 2*len(xtrain)))
    buildK(xtrain, ytrain, [x], [y], l, Kstar)
    qGP = Kstar.dot(Kyinv.dot(ztrain))
    f = qGP[1]
    return f, qGP[0]

def Pnewton(P, x, y, l, xtrain, ytrain, Kyinv, ztrain):
    Kstar = np.empty((2, 2*len(xtrain)))
    buildK(xtrain, ytrain, x, P, l, Kstar)
    pGP = Kstar.dot(Kyinv.dot(ztrain))
    f = pGP[0] - y + P
    return f

# TODO: there are still some bugs here
# def dPnewton(P, x, y, l, xtrain, ytrain, Kyinv, ztrain):
#     dKstar = build_dKdy0(xtrain, ytrain, x, P, l)
#     dpGP = dKstar.dot(Kyinv.dot(ztrain))
#     return dpGP[0] + 1

# def Pnewton_with_grad(P, x, y, l, xtrain, ytrain, Kyinv, ztrain):
#     Kstar = np.empty((2, 2*len(xtrain)))
#     buildK(xtrain, ytrain, x, P, l, Kstar)
#     dKstar = build_dKdy0(xtrain, ytrain, x, P, l)
#     pGP = Kstar.dot(Kyinv.dot(ztrain))
#     dpGP = dKstar.dot(Kyinv.dot(ztrain))
#     f = pGP[0] - y + P
#     df = dpGP[0] + 1
#     return f, [df]

def calcP(x,y, l, lp, xtrainp, ytrainp, ztrainp, Kyinvp, xtrain, ytrain, ztrain, Kyinv):
    # as P is given in an implicit relation, use newton to solve for P 
    # use the second GP on regular grid (q,p) for a first guess for P
    pgss = guessP([x], [y], lp, xtrainp, ytrainp, ztrainp, Kyinvp)
    res, r = newton(Pnewton, pgss, full_output=True, maxiter=5, disp=False,
       args = (np.array([x]), np.array([y]), l, xtrain, ytrain, Kyinv, ztrain))
    return res

def applymap(l, lp, Q0map, P0map, xtrainp, ytrainp, ztrainp, Kyinvp, xtrain, ytrain, ztrain, Kyinv):
    Nq = Q0map.shape[0]
    Np = P0map.shape[1]
    #init
    pmap = np.zeros([nm, Nq, Np])
    pmaptest = np.zeros([nm, Nq, Np])
    qmap = np.zeros([nm, Nq, Np])
    H = np.zeros([nm, Nq, Np])
    #set initial conditions
    pmap[0,:,:] = P0map
    pmaptest[0, :, :] = P0map
    qmap[0,:,:] = Q0map
    H[0,:,:] = energy([Q0map, P0map])
    
    # loop through all points on grid (q, p) and all time steps
    for i in range(0,nm-1):
        for k in range(0, Nq): 
            for lk in range(0, Np):
                # first: set new P
                pmap[i+1, k, lk] = calcP(qmap[i,k,lk], pmap[i, k, lk], l, lp, xtrainp, ytrainp, ztrainp, Kyinvp, xtrain, ytrain, ztrain, Kyinv)
                # pmap[1, k, lk] = pe0[k, lk]
                # pmap[2, k, lk] = p1[k, lk]
        for k in range(0, Nq):
            for lk in range(0, Np):
                if np.isnan(pmap[i+1, k, lk]):
                    qmap[i+1,k,lk] = np.nan
                else: 
                    # then: set new Q via calculating \Delta q and adding q
                    qmap[i+1, k, lk], dpmap = calcQ(qmap[i,k,lk], pmap[i+1,k,lk],xtrain, ytrain, l, Kyinv, ztrain)# + qmap[i, k, lk]
                    qmap[i+1, k, lk] = np.mod(qmap[i+1,k,lk]+ qmap[i, k, lk], 2.0*np.pi)
                    # testing GP on mixed grid for p -> THIS does NOT work
                    # P = p + \delta p -> as the GP is fitted for -\delta p -> additional minus
                    pmaptest[i+1, k, lk] = pmap[i,k,lk] - dpmap 
                H[i+1, k, lk] = energy([qmap[i+1, k, lk], pmap[i+1, k, lk]])
    return qmap, pmap, H, pmaptest

def dydt(y_, t):
    ydot = np.zeros([2])
    ydot[0] = y_[1] # dx/dt
    ydot[1] = -np.sin(y_[0] + np.pi)# - eps*np.sin(y_[0] - om*(t))#- 0.5*np.pi) # dpx/dt
    return ydot

def dydt_ivp(t, y):
    ydot = np.zeros([2])
    ydot[0] = y[1] # dx/dt
    ydot[1] = -np.sin(y[0]+np.pi)# - eps*np.sin(y_[0] - om*(t))#- 0.5*np.pi) # dpx/dt
    return ydot

def integrate_pendulum(q0, p0, t):
    # yint = np.zeros([len(t), 2, len(q0)]) # initial values for y
    ysint = np.zeros([len(t), 2, len(q0)]) # initial values for y

    # for k in range(len(q0)):    
    #     yint[:, :, k] = spint.odeint(dydt, [q0[k], p0[k]], t)
        # ysint[:,:, k] = intode([q0[k], p0[k]], t, dtsymp)
        # ysint[:,0, k] = np.mod(ysint[:,0, k], 2*np.pi)    
    ysint = []
    for ik in range(len(q0)):
        res_int = solve_ivp(dydt_ivp, [t[0], t[-1]], np.array((q0[ik], p0[ik])), max_step=0.001, method='DOP853')
        temp = res_int.y
        # temp[0] = np.mod(temp[0], 2*np.pi)
        ysint.append(temp)
    
    return ysint
# compute log-likelihood according to RW, p.19
def solve_cholesky(L, b):
    return solve_triangular(
        L.T, solve_triangular(L, b, lower=True, check_finite=False), 
        lower=False, check_finite=False)

# negative log-posterior
def nll_chol(hyp, x, y, build_K=build_K):
    K = np.empty((len(x), len(x)))
    build_K(x, x, hyp[:-1], K)
    Ky = K + np.abs(hyp[-1])*np.diag(np.ones(len(x)))
    L = np.linalg.cholesky(Ky)
    alpha = solve_cholesky(L, y)
    ret = 0.5*y.T.dot(alpha) + np.sum(np.log(L.diagonal()))
    #print(hyp, ret)
    return ret

def nll(hyp, x, y, neig=8, build_K=build_K):
    K = np.empty((len(x), len(x)))
    build_K(x, x, np.abs(hyp[:-1]), K)
    Ky = K + np.abs(hyp[-1])*np.diag(np.ones(len(x)))
    w, Q = eigsh(Ky, neig, tol=max(1e-6*np.abs(hyp[-1]), 1e-15))
    while np.abs(w[0]-hyp[-1])/hyp[-1] > 1e-6 and neig < len(x):
        if neig > 0.05*len(x):  # TODO: get more stringent criterion
            try: 
                return nll_chol(hyp, x, y, build_K)
                
            except:
                print('Warning! Fallback to eig solver!')
        neig =  2*neig
        w, Q = eigsh(Ky, neig, tol=max(1e-6*hyp[-1], 1e-15))

    alpha = Q.dot(np.diag(1.0/w).dot(Q.T.dot(y)))

    ret = 0.5*y.T.dot(alpha) + 0.5*(np.sum(np.log(w)) + (len(x)-neig)*np.log(np.abs(hyp[-1])))
    #print(hyp, ret)
    return ret

def plot_pendulum(ysint):
    Np = ysint.shape[3]
    Nq = ysint.shape[2]
    plt.figure()
    for kp in range(Np):
        for kq in range(Nq):
            plt.plot(ysint[:,0,kq,kp], ysint[:,1,kq,kp],'.', color = 'k', markersize = 1.0)
            plt.plot(ysint[0,0,kq,kp],ysint[0,1,kq,kp],'o', color = 'b', markersize = 4.9)
            plt.plot(ysint[-1,0,kq,kp],ysint[-1,1,kq,kp],'o', color = 'r', markersize = 4.9)
    plt.xlabel(r'$q(n{\tau})$')
    plt.ylabel(r'$p(n{\tau})$')
    
def plot_map(qmap, pmap, H, Q0map, P0map):
    plt.figure()
    for i in range(0, qmap.shape[2]):
        plt.plot(qmap[:,:,i], pmap[:,:,i], ',', markersize = 1.5)
        # plt.plot(np.mod(qmap[:,:,i]-np.pi, 2*np.pi)-np.pi, pmap[:,:,i], ',', color = 'k', markersize = 1.5)
        # plt.plot(qmap_RBF[:,:,i], pmap_RBF[:,:,i], '.', color = 'k', markersize = 1.5)
    # plt.ylim([-4, 4])
    plt.xlabel('q')
    plt.ylabel('p')
    
    plt.figure()
    for i in range(0, qmap.shape[2]):
        # i = 0
        plt.plot(H[:,:,i]/energy([Q0map[:,i], P0map[:,i]]))#/H_RBF[0,:,i])
    plt.xlabel(r'$\tau$')
    plt.ylabel('E/E_0')

def gamma(d):
    x=1.0000
    for i in range(20):
        x = x-(pow(x,d+1)-x-1)/((d+1)*pow(x,d)-1)
    return x

def bluenoise(d,n, method, qmin, qmax, pmin, pmax):

    g = gamma(d)
    alpha = np.zeros(d)                 
    for j in range(d):
        alpha[j] = pow(1/g,j+1) %1
    z = np.zeros((n, d))
    # z = []
    i = 0
    iz = 0
    while iz < n: 
        random_point = (0.5 + alpha*(i+1)) %1
        random_point[0] = random_point[0]*(np.abs(qmin)+qmax)-np.abs(qmin)
        random_point[1] = random_point[1]*(np.abs(pmin)+pmax)+pmin
        if (method == 'trapped') & (energy(random_point)/2 < 1.0):
            z[iz] = random_point
            iz = iz + 1
        elif (method == 'passing') & (energy(random_point)/2 > 1.0):
            z[iz] = random_point
            iz = iz + 1
        elif (method == 'all'):
            z[iz] = random_point
            iz = iz + 1
        i = i + 1
        #     while len(points) < num_points:
        # random_point = Point(temp[i,:])
        # if (random_point.within(poly)):
        #     points.append(random_point)
        # i = i + 1
    return z



"""
Testcases for run components
 - Interface, Preprocessor, Postprocessor
"""

from profit.config import Config
from pytest import fixture

CONFIG_FILE = 'numpy.yaml'
RUN_ID = 1
VALUE_U = -1
VALUE_V = 2
VALUE_F = 3
VALUE_T = 4


@fixture(autouse=True)
def chdir_pytest():
    from os import getcwd, chdir, path
    pytest_root_dir = getcwd()
    chdir(path.dirname(path.abspath(__file__)))
    yield
    chdir(pytest_root_dir)


def assert_wif(wif):
    assert wif.input.dtype.names == ('u', 'v')
    assert wif.output.dtype.names == ('f', 'g')
    assert wif.input.shape == ()
    assert wif.output.shape == ()
    assert wif.input['u'] == VALUE_U
    assert wif.input['v'] == VALUE_V


def test_memmap():
    from profit.run.default import MemmapInterface, MemmapRunnerInterface
    import os

    BASE_CONFIG = Config.from_file('numpy.yaml')
    MAX_IDS = BASE_CONFIG['ntrain']
    config = {'class': 'memmap'}
    try:
        MemmapRunnerInterface.handle_config(config, BASE_CONFIG)

        rif = MemmapRunnerInterface(config, MAX_IDS, BASE_CONFIG['input'], BASE_CONFIG['output'])
        rif.input[['u', 'v']][1] = VALUE_U, VALUE_V
        wif = MemmapInterface(config, RUN_ID)
        assert_wif(wif)
        wif.output['f'] = VALUE_F
        wif.time = VALUE_T
        wif.done()
        assert rif.output['f'][RUN_ID] == VALUE_F
        assert rif.internal['TIME'][RUN_ID] == VALUE_T
        assert rif.internal['DONE'][RUN_ID]
    finally:
        if 'path' in config and os.path.exists(config['path']):
            os.remove(config['path'])


def test_zeromq():
    from threading import Thread
    from time import sleep
    from profit.run.zeromq import ZeroMQInterface, ZeroMQRunnerInterface

    BASE_CONFIG = Config.from_file('numpy.yaml')
    MAX_IDS = BASE_CONFIG['ntrain']
    config = {'class': 'zeromq'}
    ZeroMQRunnerInterface.handle_config(config, BASE_CONFIG)

    def runner():
        rif = ZeroMQRunnerInterface(config, MAX_IDS, BASE_CONFIG['input'], BASE_CONFIG['output'])
        try:
            rif.input[['u', 'v']][RUN_ID] = VALUE_U, VALUE_V
            for i in range(3):
                rif.poll()
                sleep(0.5)
            assert rif.output['f'][RUN_ID] == VALUE_F
            assert rif.internal['TIME'][RUN_ID] == VALUE_T
            assert rif.internal['DONE'][RUN_ID]
        finally:
            rif.clean()

    def worker():
        wif = ZeroMQInterface(config, run_id=RUN_ID)
        assert_wif(wif)
        wif.output['f'] = VALUE_F
        wif.time = VALUE_T
        wif.done()

    rt = Thread(target=runner)
    wt = Thread(target=worker)

    rt.start()
    wt.start()
    wt.join()
    rt.join()


def test_numpytxt():
    from numpy import array
    from profit.run.default import NumpytxtPostprocessor

    BASE_CONFIG = Config.from_file('numpy.yaml')
    config = {'class': 'numpytxt', 'path': 'numpytxt.csv', 'options': {'delimiter': ','}}
    data = array([0], dtype=[('f', float, (3,)), ('g', float)])[0]

    NumpytxtPostprocessor.handle_config(config, BASE_CONFIG)
    post = NumpytxtPostprocessor(config)
    post(data)

    assert all(data['f'] == [1.4, 1.3, 1.2])
    assert data['g'] == 10
